/*-----------------------------------------------------------------------------
Modul Name : BaseAgent.cpp
Date       : 30.10.97
-------------------------------------------------------------------------------
Author     : Klaus Dorer

Description: contains the operations for a basic agent

Comments   : _
-----------------------------------------------------------------------------*/

#include "stdafx.h"
#include "Basis.h"
#include "Agentsock.h"
#include "DSetupDlg.h"
#include "DReconnect.h"
#include "BaseAgent.h"

#include <math.h>
#ifdef SHOW_TIME
	#include <time.h>
#endif


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define MAX_MSG_LENGTH 1500

/*-----------------------------------------------------------------------------
Function: CBaseAgent::CBaseAgent()
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: none
Description: constructor
-----------------------------------------------------------------------------*/
CBaseAgent::CBaseAgent()
{
	m_pActionQueue = new CStringList();
	m_bConnected = FALSE;
	m_bHadConnectionError = FALSE;
	m_iNumber = 0;
	m_strTeam = MY_TEAM;
	m_uiPortAgent = 6000;
	m_bHeadingBackward = FALSE;
}

CBaseAgent::~CBaseAgent()
{
	DisconnectAgent(FALSE);
	EmptyQueue();
	delete m_pActionQueue;
}

/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ConnectTo(BOOL bAutorun)
Date    : 14.11.97
-------------------------------------------------------------------------------
Parameter:	flag indicating, if agent has been launched by soccermonitor
Returnval: (void)
Description: tries to connect with the standard connection parameters. if not 
				 successfull, asks the user for new parameters
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ConnectTo(BOOL bAutorun)
{
	DSetupDlg Dialog;

	// ask the user for the name of the team and the server
	Dialog.m_strHandle = m_strTeam;
	Dialog.m_strServer = m_sSocketAddress;
	Dialog.m_nChannel = m_uiPortAgent;

	// user can repeat with a different address
	while(TRUE)
	{
		if (Dialog.DoModal() != IDOK)
			return FALSE;
		else
		{
			m_uiPortAgent = Dialog.m_nChannel;
			m_sSocketAddress = Dialog.m_strServer;
			m_strTeam = Dialog.m_strHandle;
		}

		// try to connect
		if (ConnectAgent())
			return TRUE;
		else if (bAutorun)
			return FALSE;
	}
}


/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ConnectAgent()
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the agent could be created and connected to a server socket
			  FALSE otherwise
Description: creates the socket of the agent and initializes it
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ConnectAgent()
{
	// is the agent already connected (should not happen) 
	if (m_bConnected)
		return FALSE;

	// create the datagram socket
	if (!Create(m_uiPortAgent, m_sSocketAddress))
	{
		// socket could not be created
		OnErrorCreatingSocket();
		m_bHadConnectionError = TRUE;
		return FALSE;
	}
	else
	{
		OnConnected();
		return TRUE;
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ReconnectTo()
Date    : 08.06.99
-------------------------------------------------------------------------------
Parameter:	none
Returnval: success or failure
Description: tries to reconnect asks the user for new parameters
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ReconnectTo()
{
	DReconnect Dialog;

	// ask the user for the name of the team and the server
	Dialog.m_strTeamname = m_strTeam;
	Dialog.m_iNumber = m_iNumber;
	Dialog.m_strServer = m_sSocketAddress;
	Dialog.m_uiPort = m_uiPortAgent;

	// user can repeat with a different address
	while(TRUE)
	{
		if (Dialog.DoModal() != IDOK)
			return FALSE;
		else
		{
			m_uiPortAgent = Dialog.m_uiPort;
			m_sSocketAddress = Dialog.m_strServer;
			m_strTeam = Dialog.m_strTeamname;
			m_iNumber = Dialog.m_iNumber;
		}

		// try to connect
		if (ReconnectAgent())
			return TRUE;
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ReconnectAgent()
Date    : 08.06.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the agent could be created and connected to a server socket
			  FALSE otherwise
Description: creates the socket of the agent and initializes it
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ReconnectAgent()
{
	// is the agent already connected (should not happen) 
	if (m_bConnected)
		return FALSE;

	// create the datagram socket
	if (!Create(m_uiPortAgent, m_sSocketAddress))
	{
		// socket could not be created
		OnErrorCreatingSocket();
		m_bHadConnectionError = TRUE;
		return FALSE;
	}
	else
	{
		OnReconnected();
		return TRUE;
	}
}

/*-----------------------------------------------------------------------------
Function: void OnConnected()
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: notification handler of a succesful socket creation
-----------------------------------------------------------------------------*/
void CBaseAgent::OnConnected()
{
	CString	message;

	// send the (init team) command
	if (m_iNumber == GOALIE_ID)
		message.Format("(init %s (goalie) (version 5.00))", (LPCTSTR) m_strTeam);
	else
		message.Format("(init %s (version 5.00))", (LPCTSTR) m_strTeam);
	SendMsg(message);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::OnReconnected()
Date    : 08.06.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: notification handler of a succesful socket creation after reconnection
-----------------------------------------------------------------------------*/
void CBaseAgent::OnReconnected()
{
	CString	message;

	if (m_iNumber > 0)
	{
		message.Format("(reconnect %s %d)", (LPCTSTR) m_strTeam, m_iNumber);
		SendMsg(message);
	}
}


/*-----------------------------------------------------------------------------
Function: void CBaseAgent::DisconnectAgent()
Date    : 16.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: empties the agent's socket and deletes it
-----------------------------------------------------------------------------*/
void CBaseAgent::DisconnectAgent(BOOL bTryingToConnect)
{
	OnDisconnect();
	if (m_bConnected || bTryingToConnect)
	{
		BYTE Buffer[2001];

		ShutDown();
		while(Receive(Buffer,2000) > 0);
		Close();
	}
	m_bConnected = FALSE;
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::OnDisconnect()
Date    : 11.05.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: sends the bye message to the server
-----------------------------------------------------------------------------*/
void CBaseAgent::OnDisconnect()
{
	SendBye();
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SendMsg(CString& strText)
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: message to be sent
Returnval: (void)
Description: the message is sent to the socket by SendTo (unconnected socket)
-----------------------------------------------------------------------------*/
void CBaseAgent::SendMsg(CString& strText)
{
	int		iLength;
	
	// send the buffer
	iLength = MySend(strText, strText.GetLength());
	if (iLength == strText.GetLength())
		OnSentMessage((LPCTSTR) strText);
	else
	{
		// an error occured
		#ifdef _DEBUG
			TRACE1("error writing to socket: %d\n", GetLastError());
		#endif
		OnErrorSendingMessage();
	}
}

/*-----------------------------------------------------------------------------
Function: void CAgentSocket::OnReceive(int nErrorCode)
Date    : 30.10.97
-------------------------------------------------------------------------------
Parameter: error code ?
Returnval: (void)
Description: called when a messege is on the socket, which can be read
-----------------------------------------------------------------------------*/
void CBaseAgent::OnReceive(int nErrorCode)
{
	char		buffer[MAX_MSG_LENGTH];
	int		iLength = 1;
	int		iPosition = 0;
	int		iNoOfMessages = 0;

#ifdef SHOW_TIME
	LARGE_INTEGER	frequency;
	LARGE_INTEGER	start;
	LARGE_INTEGER	stop;
	double			time;
	
	if (QueryPerformanceFrequency(&frequency))
		QueryPerformanceCounter(&start);
#endif

	// first call base classes method
	CAsyncSocket::OnReceive(nErrorCode);

	// read all messages from the socket
	while (iLength > 0)
	{
		iLength = ReceiveFrom(buffer, MAX_MSG_LENGTH - 1, GetAddress(), m_uiPort, 0);

		if (iLength > 0)
		{
			m_bConnected = TRUE;
			buffer[iLength] = '\0';
			OnReceivedMessage((LPCTSTR) buffer);
			iNoOfMessages++;
		}
	}

	if (iNoOfMessages == 0)
	{
		// an error occured
		#ifdef _DEBUG
			afxDump << "error reading socket\n";
		#endif
		OnErrorReceivingMessage();
	}


#ifdef SHOW_TIME
	if (QueryPerformanceCounter(&stop))
	{
		time = (double)(stop.LowPart - start.LowPart) / (double) frequency.LowPart;
		TRACE1("seetime: %4.3f\n", time);
	}
#endif

}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SetNumber(int num)
Date    : 17.03.98
-------------------------------------------------------------------------------
Parameter: number of the player
			  side the player is playing (LEFT_TO_RIGHT or RIGHT_TO_LEFT)
Returnval: (void)
Description: sets the number of the player and its position before playon
				 it is assumed, that team has been set before
-----------------------------------------------------------------------------*/
void CBaseAgent::SetNumber(int number, int side)
{ 
	// check for legal range
	if ((number < 1) || (number > PLAYERS_PER_TEAM))
	{
		#ifdef _DEBUG
			afxDump << "illegal teamnumber\n";
		#endif
		return;
	}

	m_iNumber = number; 
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Move(int x, int y)
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: x,y coordinate (in meters) where to move to before playing
Returnval: (void)
Description: following functions implement the abilities of soccer agents
				 see the soccer manual for more details
-----------------------------------------------------------------------------*/
void CBaseAgent::Move(int x, int y)
{
	// do not send messages before new port has been assigned
	if (m_bConnected)
	{
		CString	message;

		message.Format("(move %d %d)", x, y);
		SendMsg(message);

		OnMoved(x, y);
	}
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Turn(int angle, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle to turn; -180 (left) .. 180 (right)
			  flag, if this turn should not be executed but queued for next cycle
Returnval: (void)
Description: turns the agent by the passed angle. The turn is less however, if
				 the agent is moving fast
-----------------------------------------------------------------------------*/
void CBaseAgent::Turn(int angle, BOOL bQueue)
{
	CString	message;

	// angle may be -180 to 180
	message.Format("(turn %d)", angle);
	if (bQueue)
		QueueAction(message);
	else
		SendMsg(message);

	OnTurned(angle);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Dash(int power, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: power of the dash: -30 (backwards) .. 100 (forward)
			  flag, if this dash should not be executed but queued for next cycle
Returnval: (void)
Description: the agent runs in view direction (or backward) with specified power
-----------------------------------------------------------------------------*/
void CBaseAgent::Dash(int power, BOOL bQueue)
{
	CString	message;

	// power may be -30 to 100
	message.Format("(dash %d)", power);
	if (bQueue)
		QueueAction(message);
	else
	{
		SendMsg(message);
		// check if the agent is heading backward
		if (power < 0)
			m_bHeadingBackward = TRUE;
		else
			m_bHeadingBackward = FALSE;
	}

	OnDashed(power);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Kick(int power, int direction, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: power to kick: 0 .. 100
			  direction to kick in: -180 (left) ... 180 (right)
Returnval: (void)
Description: kicks the ball (if closer than 2 m) in the specified direction
				 with specified power
-----------------------------------------------------------------------------*/
void CBaseAgent::Kick(int power, int direction, BOOL bQueue)
{
	CString	message;
	
	message.Format("(kick %d %d)", power, direction);
	if (bQueue)
		QueueAction(message);
	else
		SendMsg(message);
/*
	#ifdef _DEBUG
		TRACE3("kick: power %d  dir %d  queue %d\n", power, direction, bQueue);
	#endif
*/
	OnKicked(power, direction);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Catch(int direction)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: direction to catch: -180 ... 180
Returnval: (void)
Description: a goalie can catch the ball within a rectangle that direction is
				specified
-----------------------------------------------------------------------------*/
void CBaseAgent::Catch(int direction)
{
	CString	message;
	
	// direction -180 to 180
	message.Format("(catch %d)", direction);
	SendMsg(message);

	OnCaught(direction);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Say(CString msg)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: message to say
Returnval: (void)
Description: shouts a message to other players within a restricted area
-----------------------------------------------------------------------------*/
void CBaseAgent::Say(CString msg)
{
	CString	message;

	// max 255 byte
	message.Format("(say ");
	message += msg;
	message += _T(")");
	SendMsg(message);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::ChangeView(int width, int quality)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: view angle of the agent (45, 90 or 180 degrees)
			  view quality (high, low)
Returnval: (void)
Description: changes the agents view mode. With 45 deg the agent gets perceptions
				 from the server every 75ms, with 90 every 150ms and with 180 every 300ms.
-----------------------------------------------------------------------------*/
void CBaseAgent::ChangeView(int width, int quality)
{
	CString	message;

	message =_T("(change_view ");
	// set the view angle string
	if (width == 180)
		message += _T("wide ");
	else if (width == 90)
		message += _T("normal ");
	else
		message += _T("narrow ");
	// set the view quality string
	if (quality == 0)
		message += _T("low)");
	else
		message += _T("high)");

	SendMsg(message);
	OnChangedView(width, quality);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SenseBody()
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: asks the server for information on the agent itself
-----------------------------------------------------------------------------*/
void CBaseAgent::SenseBody()
{
	if (m_bConnected)
		SendMsg(CString("(sense_body)"));
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::TurnNeck(int iDir, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle to turn neck
			  flag indicating whether the action should be queued or immediately executed (default)
Returnval: (void)
Description: turns the neck of the agent
-----------------------------------------------------------------------------*/
void CBaseAgent::TurnNeck(int iDir, BOOL bQueue)
{
	CString	message;

	// if no turn needed
	if (iDir == 0)
		return;

	// angle may be -180 to 180
	if (iDir < -180)
		iDir = -180;
	else if (iDir > 180)
		iDir = 180;

	message.Format("(turn_neck %d)", iDir);
	if (bQueue)
		QueueAction(message);
	else
		SendMsg(message);

	OnTurnedNeck(iDir);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SendBye()
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: disconnects the client from the server
-----------------------------------------------------------------------------*/
void CBaseAgent::SendBye()
{
	CString	msg("(bye)");

	SendMsg(msg);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::QueueAction(LPCTSTR action)
Date    : 25.08.98
-------------------------------------------------------------------------------
Parameter: string containing the action syntax for the server
Returnval: (void)
Description: queues the action at the end of the queue
-----------------------------------------------------------------------------*/
void CBaseAgent::QueueAction(LPCTSTR action)
{
	m_pActionQueue->AddTail(action);
}

/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ExecuteQueuedAction()
Date    : 25.08.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: True, if an action has been in the queue
Description: executes the action at the head of the list (fifo)
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ExecuteQueuedAction()
{
	if (!m_pActionQueue->IsEmpty())
	{
		SendMsg(m_pActionQueue->RemoveHead());
		return TRUE;
	}
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::EmptyQueue()
Date    : 25.08.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: removes all actions in the action queue
-----------------------------------------------------------------------------*/
void CBaseAgent::EmptyQueue()
{
	while (!m_pActionQueue->IsEmpty())
		m_pActionQueue->RemoveHead();
}
