// Basis.h : main header file for the BASIS application
//

#if !defined(AFX_BASIS_H__54CD01C9_4540_11D1_B242_0000C09B5CBE__INCLUDED_)
#define AFX_BASIS_H__54CD01C9_4540_11D1_B242_0000C09B5CBE__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
	#error include 'stdafx.h' before including this file for PCH
#endif

#include "resource.h"       // main symbols
#include "TCommandLine.h"

// defined when logging perception (needs time + space)
#ifdef _DEBUG
	#define LOG_PERCEPTION
#endif

// defined when measuring time, x out, when not
#define xSHOW_TIME
// if defined, the agent synchronizes using sense body otherwise it uses a CWnd::Timer
#define SYNCHRONIZE_ON_SENSE_BODY

//!!! types.h
#define MAX_STAMINA			3500
#define STAMINA_INCREMENT	35
#define PLAYER_SPEED_MAX	1.0
#define PLAYER_DECAY			0.4
#define PLAYER_SIZE			0.3
#define KICKABLE_MARGIN		0.7
#define KICK_POWER_RATE		0.016
#define DASH_POWER_RATE		0.006

#define BALL_SPEED_MAX		2.7
#define BALL_DECAY			0.94
#define BALL_SIZE				0.085
#define SEND_STEP				150
#define EFFORT_MIN			0.6
#define CATCHABLE_AREA_L	2.0
#define CATCHABLE_AREA_W	1.0


// constants for perceptions
#define LOW_STAMINA			1000		// point at which it costs effort

// knowledge types
#define MAES_KNOWLEDGE		1
#define GOETZ_KNOWLEDGE		2
#define DORER_KNOWLEDGE		3
#define DORER2_KNOWLEDGE	4

// player ids
#define GOALIE_ID			1
#define MAX_DEFENDER		5
#define MAX_MIDFIELDER	8

// colors
#define COLOR_MYTEAM RGB(200, 200, 20)			// yellow
#define COLOR_OTHERTEAM RGB(200, 20, 20)		//	red
#define COLOR_AGENT RGB(20, 200, 20)			// green
#define COLOR_BALL RGB(20, 20, 200)				// blue
#define COLOR_GOAL RGB(0, 0, 0)					// black
#define COLOR_FLAG RGB(200, 200, 200)			// light gray
#define COLOR_LINE RGB(255, 255, 255)			// white
#define COLOR_ANY_OBJECT RGB(40, 40, 40)		// dark gray
#define COLOR_FIELD RGB(20, 150, 20)			// green
#define COLOR_FIELD_GRAY RGB(150, 150, 150)	// lightgray
#define COLOR_VISIBLE_AREA RGB(20, 130, 20)	// green, a little lighter than above
#define COLOR_SKY RGB(30, 30, 150);				// light blue


#define DIS_VERSION	1					// version info for serialization
#define PI				3.1415926		// pi
#define D_ZERO_P		0.0000001
#define D_ZERO_N		-0.0000001



#define MAX_PLAYERS 22					// maximum number of players
#define PLAYERS_PER_TEAM 11			// actual number of players per team

#define MY_TEAM				"magmaFreiburg"			// name of my team
#define OTHER_TEAM			"TeamMaes"
#define DEFAULT_SERVER		"ultra1"
//#define DEFAULT_SERVER		"yoda"
#define DEFAULT_CHANNEL		6000				// socket channel
#define DEFAULT_MASTERCH	6001				// channel for master mode
#define ACT_SEE_RATIO		0.666666			// 100 : 150

#define LEFT_TO_RIGHT	0
#define RIGHT_TO_LEFT	1

#define MAX_PERCEPTION_LEN 1024	// maximum length of perception message
#define MAX_SAY_LEN					// maximum length of a say statement


#define NOT_EXISTING -32000	// items not existing in a perception

#define FIELD_FLAG			1
#define PENALTY_AREA_FLAG	2

// for visualizing
#define PLAYERSIZE	6			// radius of the circle representing the player
#define BALLSIZE		4
#define FLAG_SIZE		3
#define PLAYER_SIZE_DRAW		1.5	// for absolute view
#define BALL_SIZE_DRAW			1.0
#define PITCH_LENGTH 			105.0
#define PITCH_WIDTH 	 			68.0
#define PITCH_MARGIN				8.0
#define CENTER_CIRCLE_R			9.15
#define PENALTY_CIRCLE_R		9.15
#define CIRCLE_ARC				7.1
#define PENALTY_AREA_LENGTH	16.5
#define PENALTY_AREA_WIDTH		40.32
#define GOAL_AREA_LENGTH		5.5
#define GOAL_AREA_WIDTH			18.32
#define GOAL_WIDTH				14.64
#define GOAL_DEPTH				4.0
#define PENALTY_SPOT_DIST		11.0
#define CORNER_ARC_R				1.0



#define INERTIA_MOMENT	5.0

#define NO_OF_PLAYMODES		21
// playmodes 
#define BEFORE_KICK_OFF		1
#define TIME_OVER				2
#define PLAY_ON				3
#define KICK_OFF_L			4
#define KICK_OFF_R			5
#define KICK_IN_L				6
#define KICK_IN_R				7
#define FREE_KICK_L			8
#define FREE_KICK_R			9 
#define CORNER_KICK_L		10
#define CORNER_KICK_R		11
#define GOAL_KICK_L			12
#define GOAL_KICK_R			13
#define GOAL_L					14
#define GOAL_R					15
#define DROP_BALL				16
#define OFFSIDE_L				17
#define OFFSIDE_R				18
#define CATCH_L				19
#define CATCH_R				20

#define HALF_TIME				20
#define EXTEND					21
#define FOUL_L					22
#define FOUL_R					23


#define PLAYMODE_STRINGS {"",\
			"before_kick_off",\
			"time_over",\
			"play_on",\
			"kick_off_l",\
			"kick_off_r",\
			"kick_in_l",\
			"kick_in_r",\
			"free_kick_l",\
			"free_kick_r",\
			"corner_kick_l",\
			"corner_kick_r",\
			"goal_kick_l",\
			"goal_kick_r",\
			"goal_l",\
			"goal_r",\
			"drop_ball",\
			"offside_l",\
			"offside_r",\
			"goalie_catch_ball_l",\
			"goalie_catch_ball_r",\
		}

typedef enum _PlayMode {
	PM_Null,
	PM_BeforeKickOff,
	PM_TimeOver,
	PM_PlayOn,
	PM_KickOff_Left,
	PM_KickOff_Right,
	PM_KickIn_Left,
	PM_KickIn_Right,
	PM_FreeKick_Left,
	PM_FreeKick_Right,
	PM_CornerKick_Left,
	PM_CornerKick_Right,
	PM_GoalKick_Left,
	PM_GoalKick_Right,
	PM_AfterGoal_Left,
	PM_AfterGoal_Right,
	PM_Drop_Ball,
	PM_OffSide_Left,
	PM_OffSide_Right,
	PM_MAX 
} PlayMode ;

#define RAD2DEG	(180.0/PI)
#define DEG2RAD (PI/180.0)
#define Rad2Deg(a) 	((double)(a) * RAD2DEG)
#define Rad2IDeg(a)	(int)(Rad2Deg(a))
#define Deg2Rad(a)	((double)(a) * DEG2RAD)

// for blocksworld
#define GRID_SIZE		11
#define MAX_STEPS		15
#define FILENAME		"c:\\klaus\\blockklaus\\data\\result"

class TNetworkParams;
class CBasisDoc;

/////////////////////////////////////////////////////////////////////////////
// CBasisApp:
// See Basis.cpp for the implementation of this class
//

class CBasisApp : public CWinApp
{
public:
	CBasisApp();

	// for blocksworld
	void WriteResultsToFile();

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CBasisApp)
	public:
	virtual BOOL InitInstance();
	virtual BOOL OnIdle(LONG lCount);
	virtual int ExitInstance();
	//}}AFX_VIRTUAL

// Implementation

	//{{AFX_MSG(CBasisApp)
	afx_msg void OnAppAbout();
		// NOTE - the ClassWizard will add and remove member functions here.
		//    DO NOT EDIT what you see in these blocks of generated code !
	//}}AFX_MSG
	afx_msg void OnMyFileNew();
	DECLARE_MESSAGE_MAP()

private:
	TCommandLine	m_tCmdInfo;

	// for blocksworld
	CBasisDoc*			m_pDoc;
	TNetworkParams*	m_pParams;
	int					m_iLoop1;
	int					m_iLoop2;
	BOOL					m_aabResult[GRID_SIZE][GRID_SIZE];					
	int					m_iNoOfProblems;
};

extern CBasisApp theApp;

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_BASIS_H__54CD01C9_4540_11D1_B242_0000C09B5CBE__INCLUDED_)
