// BasisView.cpp : implementation of the CBasisView class
//

#include "stdafx.h"
#include "Basis.h"
#include "BasisDoc.h"

#include "BasisView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CBasisView

IMPLEMENT_DYNCREATE(CBasisView, CEditView)

BEGIN_MESSAGE_MAP(CBasisView, CEditView)
	//{{AFX_MSG_MAP(CBasisView)
	ON_WM_TIMER()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CEditView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CEditView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CEditView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CBasisView construction/destruction

CBasisView::CBasisView()
{
	m_bNoGraphics = FALSE;
	m_bTimerRunning = FALSE;
	m_iTimerID = 0;
	m_bAgentRunning = FALSE;
	m_iAgentTimerID = 0;
	m_iTimeoutTimerID = 0;
	m_bReceivedData = FALSE;
}

CBasisView::~CBasisView()
{
}

BOOL CBasisView::PreCreateWindow(CREATESTRUCT& cs)
{
	BOOL ret = CEditView::PreCreateWindow(cs);
	cs.style = AFX_WS_DEFAULT_VIEW | WS_VSCROLL | WS_HSCROLL | ES_AUTOHSCROLL |
		ES_AUTOVSCROLL | ES_MULTILINE | ES_NOHIDESEL | ES_READONLY;
	return ret;
}

/////////////////////////////////////////////////////////////////////////////
// CBasisView drawing

void CBasisView::OnDraw(CDC* pDC)
{
	// TODO: add draw code for native data here
}

/////////////////////////////////////////////////////////////////////////////
// CBasisView printing

BOOL CBasisView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CBasisView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CBasisView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CBasisView diagnostics

#ifdef _DEBUG
void CBasisView::AssertValid() const
{
	CEditView::AssertValid();
}

void CBasisView::Dump(CDumpContext& dc) const
{
	CEditView::Dump(dc);
}

CBasisDoc* CBasisView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CBasisDoc)));
	return (CBasisDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CBasisView message handlers

void CBasisView::Message(LPCTSTR lpszMessage, BOOL bFromNetwork)
{
	if (m_bNoGraphics)
		return;

	CString strTemp = lpszMessage;
	strTemp += _T("\r\n");

	if (bFromNetwork)
	{
		// if the message is from network, it is appended
		// store the string
		int len = GetWindowTextLength();
		if (len > 30000)
			// erase the middle
			GetEditCtrl().SetSel(1000,len);
		else
			// append
			GetEditCtrl().SetSel(len,len);
		GetEditCtrl().ReplaceSel(strTemp);
	}
	else
	{
		// if the message is replayed, the view gets erased
		GetEditCtrl().SetSel(0, -1);
		GetEditCtrl().ReplaceSel(strTemp);
	}
}

void CBasisView::StartConnectionTimer()
{
	if (!m_bTimerRunning)
	{
		m_bTimerRunning = TRUE;
		m_iTimerID = CONNECTION_TIMER;
		m_iTimerID = SetTimer(m_iTimerID, CONNECTION_TIMEOUT * 1000, NULL);
	}
}

void CBasisView::StopConnectionTimer()
{
	if (m_bTimerRunning)
	{
		KillTimer(m_iTimerID);
		m_bTimerRunning = FALSE;
		m_iTimerID = 0;
	}
}
void CBasisView::StartTimeoutTimer()
{
	if (m_iTimeoutTimerID == 0)
	{
		m_iTimeoutTimerID = TIMEOUT_TIMER;
		m_iTimeoutTimerID = SetTimer(m_iTimeoutTimerID, SERVER_TIMEOUT * 1000, NULL);
	}
}

void CBasisView::StopTimeoutTimer()
{
	if (m_iTimeoutTimerID != 0)
	{
		KillTimer(m_iTimeoutTimerID);
		m_iTimeoutTimerID = 0;
	}
}

void CBasisView::OnDestroy() 
{
	CEditView::OnDestroy();
	
	StopConnectionTimer();
	StopAgentTimer();
	StopTimeoutTimer();
}


void CBasisView::StartAgentTimer()
{
	#ifndef SYNCHRONIZE_ON_SENSE_BODY
		if (!m_bAgentRunning)
		{
			// start the agent acting
			m_iAgentTimerID = AGENT_TIMER;
			m_iAgentTimerID = SetTimer(m_iAgentTimerID, 100, NULL);
			m_bAgentRunning = TRUE;
		}
	#else
		m_bAgentRunning = TRUE;
	#endif
}

void CBasisView::StopAgentTimer()
{
	if (m_bAgentRunning)
	{
		KillTimer(m_iAgentTimerID);
		m_iAgentTimerID = 0;
		m_bAgentRunning = FALSE;
	}
}

void CBasisView::OnTimer(UINT nIDEvent) 
{
	if ((int)nIDEvent == m_iAgentTimerID)
	{
		// it is time for the agent to do anything
		if (m_bAgentRunning)
			GetDocument()->OnActionTimer();
	}
	
	else if ((int)nIDEvent == m_iTimeoutTimerID)
	{
		if (!m_bReceivedData)
		{
			// no message from the soccerserver, exit
			StopTimeoutTimer();
			GetDocument()->SetModifiedFlag(FALSE);
			theApp.m_pMainWnd->SendMessage(WM_CLOSE);
			return;
		}
		m_bReceivedData = FALSE;
	}

	else if ((int)nIDEvent == m_iTimerID)
	{
		// we got a timeout from connecting to server; tell the document
		StopConnectionTimer();
		GetDocument()->OnConnectionTimeout();
	}

	CEditView::OnTimer(nIDEvent);
}


