// TDorer2Network.h: interface for the TDorer2Network class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TDorer2Network_H__F8B36941_B4DB_11D1_B242_0000C09B5CBE__INCLUDED_)
#define AFX_TDorer2Network_H__F8B36941_B4DB_11D1_B242_0000C09B5CBE__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "TBehaviorNetwork.h"

class TCompetence;
class TNetworkView;
class TRuleView;
class TGoal;
class TPercept;
class TResource;
class TResourceFactory;

/*-----------------------------------------------------------------------------
Class   : TDorer2Network
Date    : 08.01.98
-------------------------------------------------------------------------------
Base Class: CObject
Description: represents the knowledge of an agent containing the three levels
				 of cognition: 
				 cognitivly inpenetrable perception and action
				 competence modules
				 goals and goal management
				 implementation of Maes Agent Network Architecture (ANA) extended
				 to use multivalued propositions
-----------------------------------------------------------------------------*/
class DLL_CLASS TDorer2Network : public TBehaviorNetwork
{
	public:
		TDorer2Network(TNetworkView* pNView, TRuleView* pRView);
		virtual ~TDorer2Network();

		// public methods
		virtual void	Initialize();
		virtual void	ParamsDialog();
		virtual void	UpdatePerception(BOOL bFromNetwork);
		virtual BOOL	ParseCompetenceRule(TCompetence* pCompetence, LPCTSTR rule);
		virtual BOOL	DoActionSelection();
		virtual void	DrawNetwork(CDC* pDC, TScreenParamsNetwork* pScreen);
		virtual void	SetNormCoordinates();
		virtual void	SetCoordinates(TScreenParamsNetwork* pScreen);
		virtual void	ReduceThreshold();

		void				AddResource(LPCTSTR name, CObject* pObj, PCallBackPerception pFn, int iAmount);
		void				SetUseFuzzyValues(BOOL bVal);
		BOOL				GetUseFuzzyValues() const;

	protected:
		virtual TCompetence*	CreateNewCompetence();
		virtual TGoal*			CreateNewGoal();
		
		// for blocksworld
		virtual BOOL	ParseGoals(TParseString &str);			

		CObList*				m_pResourceList;
		TResourceFactory*	m_pResourceFactory;	

	private:
		void	ResetResourceUsage();
};

#endif // !defined(AFX_TDorer2Network_H__F8B36941_B4DB_11D1_B242_0000C09B5CBE__INCLUDED_)
