// TNetworkParams.h: interface for the TNetworkParams class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TNETWORKPARAMS_H__9A424901_F610_11D1_A16C_006008E88C8E__INCLUDED_)
#define AFX_TNETWORKPARAMS_H__9A424901_F610_11D1_A16C_006008E88C8E__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000


#define GAMMA						1.5
#define DELTA						1.0
#define PHI							0.4
#define THETA						2.0
#define THRESHOLD_REDUCTION	0.2	// 10 % of THETA
#define NETWORK_PI				1.8 
#define MAX_ACTIVATION			4.0

#define BETA						0.5
#define MAX_ACTIVATION_SIGMOID 1.0
#define SIGMA						0.5	// must be MAX_ACTIVATION / 2
#define GAIN						5.0	// steepness of the sigmoid function
#define GAMMA_SIGMOID			0.5
#define DELTA_SIGMOID			0.2
#define PHI_SIGMOID				0.1
#define THETA_SIGMOID			0.4
#define TH_REDUCTION_SIGMOID	0.1	// 10 % of THETA


/*-----------------------------------------------------------------------------
Class   : TNetworkParams  
Date    : 28.05.98
-------------------------------------------------------------------------------
Base Class: _

Description: contains the parameters for behavior networks (Maes and Dorer)
-----------------------------------------------------------------------------*/
class DLL_CLASS TNetworkParams  
{
	public:
		TNetworkParams();
		TNetworkParams(TNetworkParams* params);
		virtual ~TNetworkParams();

		void		SetNetworkParams(TNetworkParams* params);

		// get functions
		double	GetPhi(void) const;
		double	GetGamma(void) const;
		double	GetDelta(void) const;
		double	GetPhiByGamma(void) const;
		double	GetDeltaByGamma(void) const;
		double	GetNetworkPi(void) const;
		double	GetTheta(void) const;
		double	GetThetaReduction(void) const;
		double	GetSigma(void) const;
		double	GetGain(void) const;
		double	GetBeta(void) const;
		double	GetFuzzyRound(void) const;

		double	GetPhiTimesGamma(void) const;
		double	GetGammaSquare(void) const;
		double	GetDeltaTimesGamma(void) const;

		// set functions
		void		SetPhi(double val);
		void		SetGamma(double val);
		void		SetDelta(double val);
		void		SetNetworkPi(double val);
		void		SetTheta(double val);
		void		SetThetaReduction(double val);
		void		SetSigma(double val);
		void		SetGain(double val);
		void		SetBeta(double val);
		void		SetFuzzyRound(double val);

	private:
		double	m_dPhi;
		double	m_dGamma;
		double	m_dDelta;
		double	m_dTheta;
		double	m_dThetaReduction;
		double	m_dSigma;
		double	m_dGain;
		double	m_dBeta;

		double	m_dFuzzyRound;

		double	m_dPhiByGamma;
		double	m_dDeltaByGamma;
		double	m_dPhiTimesGamma;
		double	m_dGammaSquare;
		double	m_dDeltaTimesGamma;
};

#endif // !defined(AFX_TNETWORKPARAMS_H__9A424901_F610_11D1_A16C_006008E88C8E__INCLUDED_)
