// TCommandLine.cpp: implementation of the TCommandLine class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Basis.h"
#include "TCommandLine.h"

#include <stdlib.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TCommandLine::TCommandLine()
{
	m_bAutorun = FALSE;
	m_bHasRuleFile = FALSE;
	m_bHasTeamname = FALSE;
	m_bHasServerAddress = FALSE;
	m_bHasMonitorAddress = FALSE;
	m_bHasPort = FALSE;
	m_bHasType = FALSE;
	m_bHasGoalTracking = FALSE;
	m_bHasFuzzy = FALSE;
	m_bHasAdaptActivationArea = FALSE;
	m_bHasPhi = FALSE;
	m_bHasGamma = FALSE;
	m_bHasDelta = FALSE;
	m_bHasPi = FALSE;
	m_bHasTheta = FALSE;	
	m_bHasThetaReduction = FALSE;	
	m_bHasFuzzyRound = FALSE;

	m_iAutorunID = -1;
	m_iPlayerID = -1;
	m_uiMonitorPort = -1;
	m_iPort = -1;
	m_iType = -1;
	m_dPhi = NO_DOUBLE;
	m_dGamma = NO_DOUBLE;
	m_dDelta = NO_DOUBLE;
	m_dPi = NO_DOUBLE;
	m_dTheta = NO_DOUBLE;
	m_dThetaReduction = NO_DOUBLE;
	m_dFuzzyRound = NO_DOUBLE;
}

TCommandLine::~TCommandLine()
{
}

void TCommandLine::ParseParam( LPCTSTR lpszParam, BOOL bFlag, BOOL bLast )
{
	double	value;
	char*		endptr = NULL;

	// check the name and parameter
	if (bFlag)
	{
		// it is a paramter name
		if (lstrcmpA(lpszParam, "team") == 0)
			m_bHasTeamname = TRUE;
		else if (lstrcmpA(lpszParam, "server") == 0)
			m_bHasServerAddress = TRUE;
		else if (lstrcmpA(lpszParam, "lserver") == 0)
			m_bHasMonitorAddress = TRUE;
		else if (lstrcmpA(lpszParam, "port") == 0)
			m_bHasPort = TRUE;
		else if (lstrcmpA(lpszParam, "a") == 0)
			m_bAutorun = TRUE;
		else if (lstrcmpA(lpszParam, "rule") == 0)
			m_bHasRuleFile = TRUE;
		else if (lstrcmpA(lpszParam, "type") == 0)
			m_bHasType = TRUE;
		else if (lstrcmpA(lpszParam, "gt") == 0)
			m_bHasGoalTracking = TRUE;
		else if (lstrcmpA(lpszParam, "fz") == 0)
			m_bHasFuzzy = TRUE;
		else if (lstrcmpA(lpszParam, "aaa") == 0)
			m_bHasAdaptActivationArea = TRUE;
		else if (lstrcmpA(lpszParam, "phi") == 0)
			m_bHasPhi = TRUE;
		else if (lstrcmpA(lpszParam, "ga") == 0)
			m_bHasGamma = TRUE;
		else if (lstrcmpA(lpszParam, "de") == 0)
			m_bHasDelta = TRUE;
		else if (lstrcmpA(lpszParam, "pi") == 0)
			m_bHasPi = TRUE;
		else if (lstrcmpA(lpszParam, "th") == 0)
			m_bHasTheta = TRUE;
		else if (lstrcmpA(lpszParam, "thR") == 0)
			m_bHasThetaReduction = TRUE;
		else if (lstrcmpA(lpszParam, "fzR") == 0)
			m_bHasFuzzyRound = TRUE;
		else
			CCommandLineInfo::ParseParam(lpszParam, bFlag, bLast);
	}
	else 
	{
		// it is not the name part of a parameter, but the value
		if (m_bHasTeamname && (m_strTeamname.IsEmpty()))
			// teamnameparameter
			m_strTeamname = lpszParam;
		else if (m_bHasRuleFile && (m_strRuleFile.IsEmpty()))
			// filename of rule file
			m_strRuleFile = lpszParam;
		else if (m_bHasServerAddress && (m_strServerAddress.IsEmpty()))
			// server address
			m_strServerAddress = lpszParam;
		else if (m_bHasMonitorAddress && (m_strMonitorAddress.IsEmpty()))
			// monitor address
			m_strMonitorAddress = lpszParam;
		else if (m_bHasMonitorAddress && (m_uiMonitorPort == -1))
			// monitor port
			m_uiMonitorPort = atoi(lpszParam);
		else if (m_bHasPort && (m_iPort == -1))
			// server port
			m_iPort = atoi(lpszParam);
		else if (m_bAutorun && (m_iAutorunID == -1))
			// autorun id following
			m_iAutorunID = atoi(lpszParam);
		else if (m_bAutorun && (m_iPlayerID == -1))
			// player id following
			m_iPlayerID = atoi(lpszParam);
		else if (m_bHasType && (m_iType == -1))
		{
			// knowledge type parameter
			if (lstrcmpi(lpszParam, "Maes") == 0)
				m_iType = MAES_KNOWLEDGE;
			else if (lstrcmpi(lpszParam, "Goetz") == 0)
				m_iType = GOETZ_KNOWLEDGE;
			else if (lstrcmpi(lpszParam, "Dorer") == 0)
				m_iType = DORER_KNOWLEDGE;
			else if (lstrcmpi(lpszParam, "Dorer2") == 0)
				m_iType = DORER2_KNOWLEDGE;
		}
		else
		{
			// it should be a parameter value
			value = strtod( lpszParam, &endptr);
			if (m_bHasPhi && (m_dPhi == NO_DOUBLE))
				m_dPhi = value;
			else if (m_bHasGamma && (m_dGamma == NO_DOUBLE))
				m_dGamma = value;
			else if (m_bHasDelta && (m_dDelta == NO_DOUBLE))
				m_dDelta = value;
			else if (m_bHasPi && (m_dPi == NO_DOUBLE))
				m_dPi = value;
			else if (m_bHasTheta && (m_dTheta == NO_DOUBLE))
				m_dTheta = value;
			else if (m_bHasThetaReduction && (m_dThetaReduction == NO_DOUBLE))
				m_dThetaReduction = value;
			else if (m_bHasFuzzyRound && (m_dFuzzyRound == NO_DOUBLE))
				m_dFuzzyRound = value;
			else
				CCommandLineInfo::ParseParam(lpszParam, bFlag, bLast);
		}
	}
}
