// TDoingAgent.h: interface for the TDoingAgent class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TDOINGAGENT_H__B6B1C1C2_8CBF_11D1_B242_0000C09B5CBE__INCLUDED_)
#define AFX_TDOINGAGENT_H__B6B1C1C2_8CBF_11D1_B242_0000C09B5CBE__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "TViewingAgent.h"

#ifdef _DEBUG
	// define to see some perception variables in output window
	#define xSHOW_DOING_VAR
#endif

class CBasisDoc;
class CBasisView;
class TPerceptionView;
class TRuleView;
class TNetworkView;
class TNetworkParams;
class TAct;
class TPercept;
class TVector;

/*-----------------------------------------------------------------------------
Class   : TDoing Agent
Date    : 14.01.98
-------------------------------------------------------------------------------
Base Class: TViewingAgent
Description: adds the knowledge things to an agent
-----------------------------------------------------------------------------*/
class TDoingAgent : public TViewingAgent  
{
	public:
		TDoingAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
						, TNetworkView* pNView, TRuleView* pKView);
		virtual ~TDoingAgent();

		// some more complex actions
		void ComplexDash(double dDistance, int iDirection, int iPower = 0, BOOL bMindBall = FALSE);
		void ComplexKick(int iPower, int iDirection, bool bQueue = false);
		void ComplexDashAndKick(int iPower, int iDirection);
		BOOL ComplexDashTo(TVector& pos, int power = 100, int dist = 3, BOOL bMindBall = FALSE);
		BOOL ComplexDashTo(TVisibleObject* pObj, int power = 100, int dist = 3, BOOL bMindBall = FALSE);
		void ComplexKickTo(TVector& pos, BOOL bStopThere = FALSE);
		void ComplexKickTo(TVisibleObject* pObj, BOOL bStopThere = FALSE);
		
		virtual void		TurnNeck(int iDir, BOOL bQueue = FALSE);
		virtual void		SetKnowledgeType(int type);
		virtual void		OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork = TRUE);
		void					OnAgentReady();
		void					OnReceivedSee();
		void					OnReceivedLastSenseBody();
		void					OnTimer();

		void					SetGoalTrackingOn(BOOL bVal);
		BOOL					GetGoalTrackingOn() const;
		void					SetUseFuzzyValues(BOOL val);
		BOOL					GetUseFuzzyValues() const; 
		void					SetFuzzyRound(double val);
		void					SetAdaptActivationAreaOn(BOOL bVal);
		BOOL					GetAdaptActivationAreaOn() const;

		void					SetCombo(CComboBox* pCombo, BOOL bWhich);
		CString				GetActionPerformedToString();
		void					ReceivedTerminate();
		CString				GetResults();
		CString				GetResultsInitial() const;
		BOOL					IsAutorun() const;

		// --------------------------------------------------
		// action functions (sorted alphabetically !!)

		void	CreateActions();
		void	DummyAction(double activation) {};
		
		void	DoANothing(double activation);
		void	DoCatchBall(double activation);
		void	DoCheckOffside(double activation);
		void	DoKickAwayBall(double activation);
		void	DoKickGoal(double activation);
		void	DoLookStraight(double activation);
		void	DoLookToBall(double activation);
		void	DoMindBall(double activation);
		void	DoMoveToHomePosition(double activation);
		void	DoPassBall(double activation);
		void	DoPushBackward(double activation);
		void	DoPushForward(double activation);
		void	DoRelax(double activation);
		void	DoRetreat(double activation);
		void	DoRunFree(double activation);
		void	DoRunForwardToBall(double activation);
		void	DoRunToBall(double activation);
		void	DoRunToHomePosition(double activation);
		void	DoRunToOtherGoal(double activation);
		void	DoRunToOwnGoal(double activation);
		void	DoRunWithBall(double activation);
		void	DoSayPosition(double activation = 1.0);
		void	DoSearchBall(double activation = 1.0);
		void	DoSearchTeammate(double activation);
		void	DoTurnAroundBallForKickIn(double activation);
//		void	BallDribbleToEnemyGoal(bool dribbleOnTheLeftSide);
//		bool	BallRotateLeftOrRight(TDirection deltaBallStopAngle);
//		void	BallRotateAroundPlayer(TDirection deltaBallStopAngle, bool rotateRight);
		// --------------------------------------------------

		double	GetAvgDistance() const { return m_dAvgDistance; };

		// for blocksworld
		BOOL			IsBlocksworld();
		BOOL			PerformCycle(TNetworkParams* params, int iNoOfCycles);
		void			StopAgent();

	protected:

		BOOL					DoReflexSelection();

		CBasisDoc*			m_pDoc;
		BOOL					m_bAgentIsReady;
		BOOL					m_bReceivedSee;
		BOOL					m_bSelectedReflex;
		int					m_iSenseBodyCount;
		// stores the kickangle of the ball rotating around the player
		int					m_iKickAngle;

		// for statistics
		double	m_dAvgDistance;
		long		m_lNoOfRelaxes;


	private:
		void		AccountBallSpeed(int& power, TDirection& dir);
		void		DribbleCounterclockwise();
		void		DribbleClockwise();
};

#endif // !defined(AFX_TDOINGAGENT_H__B6B1C1C2_8CBF_11D1_B242_0000C09B5CBE__INCLUDED_)
