// TPerception.cpp: implementation of the TPerception class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "TPerceptionView.h"
#include "TViewingAgent.h"
#include "TThisPlayer.h"
#include "TScreenParams.h"
#include "TPerception.h"

#include <math.h>
#include <stdio.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TPerception::TPerception(TViewingAgent* pAgent)
{
	m_pAgent = pAgent;
	CreatePerception();
	m_iSide = LEFT_TO_RIGHT;
}

TPerception::TPerception(TViewingAgent* pAgent, TPerceptionView* pView)
{
	m_iSide = LEFT_TO_RIGHT;
	m_pAgent = pAgent;
	m_pView = pView;
	m_pView->SetPerception(this);
	CreatePerception();
}

void TPerception::CreatePerception()
{
	int	i;
	double	dFlagX[NO_OF_FLAGS] = FLAG_POSITIONS_X;
	double	dFlagY[NO_OF_FLAGS] = FLAG_POSITIONS_Y;

	// create the needed objects
	// players
	m_pThisPlayer = new TThisPlayer();
	m_pThisPlayer->SetPlayer(0, 0.0, 0, 0.0, 0, 0, 0, TRUE, m_iSide, 0);
	for (i = 0; i < MAX_PLAYERS; i++)
		m_pVisiblePlayer[i] = new TPlayer();

	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i] = new TPlayer();
		m_pOtherTeam[i] = new TPlayer();
	}

	// goals (the x coordinates may be swaped later, when it is clear
	// which team the player is
	m_pOwnGoal = new TSoccerGoal(-52.5, 0.0);
	m_pOtherGoal = new TSoccerGoal(52.5, 0.0);
	m_pOwn11mPoint = new TFlag(-41.5, 0.0);
	m_pOther11mPoint = new TFlag(41.5, 0.0);

	// flags
	for (i = 0; i < NO_OF_FLAGS; i++)
		m_pFlag[i] = new TFlag(dFlagX[i], dFlagY[i]);

	// lines
	for (i = 0; i < 5; i++)
		m_pLine[i] = new TLine();

	// ball
	m_pBall = new TBall();
	
	for(i = 0; i < MAX_FUTURE_POSITIONS; i++)
		m_pFutureBallPosition[i] = new TVector(0.0, 0.0);

	m_bWaitForLastSenseBody = FALSE;
	m_bReceivedSee = FALSE;
	m_iNoOfPlayers = 0;
	m_iPlayMode = BEFORE_KICK_OFF;
	m_iViewQuality = 1;
	m_iViewAngle = 90;
	m_iSeeTime = 0;
	m_iHearTime = 0;
	m_iSenseTime = 0;
	m_iLastCatchTime = 0;
	m_bTimeProceeding = FALSE;
	m_bGoalieCaught = FALSE;
	m_bKnowNotInOffside = TRUE;
	m_dInOffside = 0.0;
	m_iCheckOffsideCounter = 0;

	SetStamina(MAX_STAMINA);
	m_dEffort = 1.0;
	m_dSpeed = 0.0;
	m_iNoOfKick = 0;
	m_iNoOfDash = 0;
	m_iNoOfTurn = 0;
	m_iNoOfSay = 0;
	m_iNoOfTurnNeck = 0;
	m_iHeadAngle = 0;

	m_iExpectedNoOfKick = 0;
	m_iExpectedNoOfDash = 0;
	m_iExpectedNoOfTurn = 0;
	m_iExpectedNoOfSay = 0;
	m_iExpectedNoOfTurnNeck = 0;
	m_iTurnAngle = 0;
	m_iTurnAngleNeck = 0;

	m_pTeammateAtBall = NULL;
	m_pTeammateAtMe = NULL;
	m_pFrontmostTeammate = NULL;
	m_pBackmostTeammate = NULL;
	m_p2BackmostTeammate = NULL;
	m_pFreeTeammate = NULL;

	m_pOpponentAtBall = NULL;
	m_pOpponentAtMe = NULL;
	m_pBackmostOpponent = NULL;
	m_p2BackmostOpponent = NULL;
}

TPerception::~TPerception()
{
	int	i;

	// delete visible objects
	delete m_pThisPlayer;
	for (i = 0; i < MAX_PLAYERS; i++)
		delete m_pVisiblePlayer[i];

	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		delete m_pOwnTeam[i];
		delete m_pOtherTeam[i];
	}

	// delete the flags
	for (i = 0; i < NO_OF_FLAGS; i++)
		delete m_pFlag[i];
	
	// delete the lines
	for (i = 0; i < 5; i++)
		delete m_pLine[i];

	delete m_pOwn11mPoint;
	delete m_pOther11mPoint;
	delete m_pOwnGoal;
	delete m_pOtherGoal;
	delete m_pBall;
	for(i = 0; i < MAX_FUTURE_POSITIONS; i++)
		delete m_pFutureBallPosition[i];
}

/*-----------------------------------------------------------------------------
Function: void TPerception::Initialize()
Date    : 6.5.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: initialize does reset the time the agent has seen the ball or any player
				 this is needed for the absolute view of perceptions and mind
-----------------------------------------------------------------------------*/
void TPerception::Initialize()
{
	int	i;

	// initialize the ball
	m_pBall->Initialize();

	// initializes the players
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i]->Initialize();
		m_pOtherTeam[i]->Initialize();
	}	

	// empty the message pane
}

/*-----------------------------------------------------------------------------
Function: void TPerception::InvalidateAll()
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: invalidates all visible objects. Invalidate means, they are not drawn
				 if they are not perceived in this cycle
-----------------------------------------------------------------------------*/
void TPerception::InvalidateAll()
{
	int	i;

	// ThisPlayer does not get invalidated, because it is always drawn
	// Invalidate Players
	m_iNoOfPlayers = 0;

	// invalidate remembered players
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i]->Invalidate();
		m_pOtherTeam[i]->Invalidate();
	}

	// and other objects
	for (i = 0; i < NO_OF_FLAGS; i++)
		m_pFlag[i]->Invalidate();

	for (i = 0; i < 5; i++)
		m_pLine[i]->Invalidate();

	m_pOwnGoal->Invalidate();
	m_pOtherGoal->Invalidate();
	m_pBall->Invalidate();
}


/*-----------------------------------------------------------------------------
Function: void TPerception::UpdatePerception()
Date    : 23.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called each 100 ms (when an action should be selected) to update
				 the moveable objects positions, if no perception has been received
				 since the last action selection
-----------------------------------------------------------------------------*/
void TPerception::UpdatePerception()
{
	int	i;
	int	iDataAge = m_iSenseTime - m_iSeeTime;

	// update all moveable objects
	m_pBall->UpdatePosition(iDataAge, m_iSeeTime);
		
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i]->UpdatePosition(iDataAge, m_iSeeTime);
		m_pOtherTeam[i]->UpdatePosition(iDataAge, m_iSeeTime);
	}

	// update the agents own position and direction according to the current speed.
	m_pThisPlayer->UpdatePlayerCoordinates(iDataAge, m_dSpeed, m_pAgent->IsDashingBackward());
}

/*-----------------------------------------------------------------------------
Function: void TPerception::DrawPerception(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
			  parameters for drawing
Returnval: (void)
Description: draws the last received perception
-----------------------------------------------------------------------------*/
void TPerception::DrawPerception(CDC* pDC, TScreenParams *tScreen)
{
	int		i;

	tScreen->SetViewAngle(m_iViewAngle);
	if (tScreen->IsRelativeView())
	{
		// draw from the players perspective
		// sort the objects depending on their distance
		CObList	drawQueue;
		
		Insert(drawQueue, m_pOwnGoal);
		Insert(drawQueue, m_pOtherGoal);

		// and the flags
		for (i = 0; i < NO_OF_FLAGS; i++)
			Insert(drawQueue, m_pFlag[i]);

		// draw the directly perceived line tags
		for (i = 0; i < 5; i++)
			Insert(drawQueue, m_pLine[i]);

		// draw the visible players
		for (i = 0; i < m_iNoOfPlayers; i++)
			Insert(drawQueue, m_pVisiblePlayer[i]);

		// draw the ball
		Insert(drawQueue, m_pBall);

		// draw the visible area
		m_pThisPlayer->DrawVisibleArea(pDC, tScreen);

		// and draw them
		POSITION pos = drawQueue.GetHeadPosition();
		while (pos != NULL)
			((TVisibleObject*) drawQueue.GetNext(pos))->DrawObject(pDC, tScreen);
	}
	else if (tScreen->IsTopView())
	{
		// draw the visible area
		m_pThisPlayer->DrawVisibleArea(pDC, tScreen);

		// draw the goals
		m_pOwnGoal->DrawGoal(pDC, tScreen, GetObjectDistance(m_pOwnGoal));
		m_pOtherGoal->DrawGoal(pDC, tScreen, GetObjectDistance(m_pOtherGoal));

		// and the flags
		for (i = 0; i < NO_OF_FLAGS; i++)
			m_pFlag[i]->DrawObject(pDC, tScreen);

		// draw the directly perceived line tags
		for (i = 0; i < 5; i++)
			m_pLine[i]->DrawObject(pDC, tScreen);

		// draw the visible players
		for (i = 0; i < m_iNoOfPlayers; i++)
			m_pVisiblePlayer[i]->DrawObject(pDC, tScreen);

		// draw this player
		m_pThisPlayer->DrawObject(pDC, tScreen);

		// draw the ball
		m_pBall->DrawObject(pDC, tScreen);
	}
	else
	{
		// set the current time (for absolute view)
		tScreen->SetTime(m_iSeeTime);

		// draw the visible area
		m_pThisPlayer->DrawVisibleArea(pDC, tScreen);

		// and the flags
		for (i = 0; i < NO_OF_FLAGS; i++)
			m_pFlag[i]->DrawObject(pDC, tScreen);

		// draw the players of the agents mind
		for (i = 0; i < PLAYERS_PER_TEAM; i++)
		{
			if (!m_pOwnTeam[i]->IsThisAgent())			
				m_pOwnTeam[i]->DrawObject(pDC, tScreen);
			m_pOtherTeam[i]->DrawObject(pDC, tScreen);
		}

		// draw this player
		m_pThisPlayer->DrawObject(pDC, tScreen);

		// draw future positions of the ball
		if (tScreen->GetDrawFutureBall())
		{
			TBall		fBall;
			TPlayer	fAgent;
			BOOL		bDrawText = tScreen->GetDrawDegreeOfFreeness();
			TVector	position;
			int		timeAhead;

			tScreen->SetDrawDegreeOfFreeness(FALSE);
			for (i = 0; i < MAX_FUTURE_POSITIONS / 2 - 1; i++)
			{
				timeAhead = i * 2 + 2;
				fBall.Invalidate();
				fBall.SetOnceSeen();
				fBall.SetInMind();
				fBall.SetPosition(*m_pFutureBallPosition[timeAhead]);
				fBall.DrawObject(pDC, tScreen);
				m_pThisPlayer->GetFuturePosition(timeAhead, position);
				fAgent.Invalidate();
				fAgent.SetOnceSeen();
				fAgent.SetInMind();
				fAgent.SetPosition(position);
				fAgent.DrawObject(pDC, tScreen);
			}
			tScreen->SetDrawDegreeOfFreeness(bDrawText);
		}

		// draw the ball
		m_pBall->DrawObject(pDC, tScreen);
	}
}

/*-----------------------------------------------------------------------------
Function: void TPerception::Insert(CObList &list, TVisibleObject* obj)
Date    : 12.02.98
-------------------------------------------------------------------------------
Parameter: reference to the draw list
			  pointer to the object, that has to be inserted
Returnval: (void)
Description: inserts sorted, by distance of the object, descending. used for drawing
-----------------------------------------------------------------------------*/
void TPerception::Insert(CObList &list, TVisibleObject* obj)
{
	TVisibleObject*	compare;
	
	// only valid objects are drawn
	if (!obj->IsValid())
		return;

	POSITION oldPos;
	POSITION	pos = list.GetHeadPosition();
	while(pos != NULL)
	{
		oldPos = pos;
		compare = (TVisibleObject*) list.GetNext(pos);
		if (obj->GetDistance() >= compare->GetDistance())
		{
			list.InsertBefore(oldPos, obj);
			return;
		}
	}

	// it is the nearest object so far
	list.AddTail(obj);
}

/*-----------------------------------------------------------------------------
Function: void TPerception::SetSide(int side) 
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: side on which the player is playing
Returnval: (void)
Description: 
-----------------------------------------------------------------------------*/
void TPerception::SetSide(int side) 
{ 
	m_iSide = side; 
	m_pThisPlayer->SetTeam(TRUE, side); 
}

/*-----------------------------------------------------------------------------
Function: void TPerception::PlayStep(LPCTSTR msg)
Date    : 19.05.98
-------------------------------------------------------------------------------
Parameter: pointer to a string containing the message (perception string)
Returnval: (void)
Description: plays a step of the stored logfile, by calling the agents onreceivedmessage
-----------------------------------------------------------------------------*/
void TPerception::PlayStep(CStringList* pList)
{
	POSITION pos = pList->GetHeadPosition();
	CString	temp;

	// parse all messages of this cycle
	while (pos != NULL)
	{
		temp = pList->GetNext(pos);
		m_pAgent->OnReceivedMessage(temp, FALSE);
	}

	// print all other messages of this cycle
	pos = pList->GetHeadPosition();
	while (pos != NULL)
	{
		temp = pList->GetNext(pos);
		m_pAgent->OnSentMessage(temp);
	}
}

/*-----------------------------------------------------------------------------
Function: int	TPerception::SwitchViewAngle(void)
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: the new view angle
Description: switches the internal representation of the viewing angle from 
				 180 degrees to 90 degrees to 45 degrees
-----------------------------------------------------------------------------*/
int	TPerception::SwitchViewAngle()
{ 
	if (m_iViewAngle == 180)
		m_iViewAngle = 90;
	else if (m_iViewAngle == 90)
		m_iViewAngle = 45;
	else
		m_iViewAngle = 180;
	
	return m_iViewAngle;
}	

/*-----------------------------------------------------------------------------
Function: void	TPerception::SwitchViewQuality(void)
Date    : 19.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: switches the view quality from high to low and back to high
-----------------------------------------------------------------------------*/
int TPerception::SwitchViewQuality()
{
	if (m_iViewQuality == 0)
		m_iViewQuality = 1;
	else
		m_iViewQuality = 0;

	return m_iViewQuality;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::NewMessage(LPCTSTR str)
Date    : 05.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the message
Returnval: (void)
Description: calls the views function for new message,which stores it and calls 
				 ParseMessage
-----------------------------------------------------------------------------*/
void TPerception::NewMessage(LPCTSTR str)
{
	m_pView->NewMessage(str);
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseMessage(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: string containing the perception (max 1024 char)
Returnval: TRUE, if no parsing errors occured
Description: distinguishes between see and hear messages and calls the appropriate
				 specialized parsing functions
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseMessage(LPCTSTR str)
{
	switch(str[1])
	{
		case 'h':
			// something to hear
			return ParseHear(str);
		
		case 's':
			// something to see or sense body
			if (str[3] == 'e')
			{
				// a see message
				if (!ParseSee(str))
					return FALSE;

				// there is a working connection
				if (!m_bReceivedSee)
				{
					m_bReceivedSee = TRUE;
					m_pAgent->OnReceivedSee();
				}

				// agent got a see
				m_pAgent->SetWaitForNextSee(FALSE);

				// get the absoulte localization of the player
				Localize();

				// check if we received a see that does not take our last turn into account
				if ((m_iNoOfTurn < m_iExpectedNoOfTurn) && (m_iSeeTime == m_iSenseTime))
					UpdatePerceptionAfterTurn(m_iTurnAngle);
					
				// same for turn_neck
				if ((m_iNoOfTurnNeck < m_iExpectedNoOfTurnNeck) && (m_iSeeTime == m_iSenseTime))
					UpdatePerceptionAfterTurnNeck(m_iTurnAngleNeck);
				
				// and for dash
				if ((m_iNoOfDash < m_iExpectedNoOfDash) && (m_iSeeTime == m_iSenseTime))
					UpdatePerceptionAfterDash(m_iDashPower);

				// calculate the absolute positions of the other moveable objects perceived
				GlobalMap();

				// for kick it is done after GlobalMap, because it changes speed vector
				// of the ball which would be overwritten by GlobalMap
				if ((m_iNoOfKick < m_iExpectedNoOfKick) && (m_iSeeTime == m_iSenseTime)
					&& m_pBall->IsValid())
					UpdatePerceptionAfterKick(m_iKickPower, m_iKickDirection);
					
				// try to match the perceived players to remembered players
				MatchPlayers();

				// remove objects not seen, but should be seen
				CheckVisualPlausibility();

				// calculate the objects functions in the world such as opponentNearBall
				CalculateFunctionalObjects();
			}

			else
				// a sense body information
				return ParseSenseBody(str);
			
			break;

		case 'i':
			// message at init
			ParseInit(str);
			break;

		case 'r':
			// message at reconnect
			return TRUE;
	}
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseSee(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: reference of the string to parse
			  reference to the current position
Returnval: TRUE, if no parsing errors occured
Description: parses a see perception
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseSee(LPCTSTR str)
{
	TParseString	msg(str);
	TParseString	strObject;
	int				time;

	// delete previously seen players
	InvalidateAll();

	// get the time
	msg.SetPosition(5);
	time = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing seetime  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// if the time is running, then the game is on
//	if ((time > m_iSeeTime) && (time > m_iHearTime + 1))
//		m_iPlayMode = PLAY_ON;

	m_iSeeTime = time;
	m_pThisPlayer->SetTime(m_iSeeTime);

	// get the visible objects
	while (msg.ParseBracket(strObject))
	{
		if (!ParseObject(strObject))
		{
			#ifdef _DEBUG
				CString error;
				error.Format(" : error parsing object info at position %d !\n", strObject.GetPosition());
				afxDump << error << strObject;
			#endif
		}
	};

	// may be not true to return True because of parsebracket
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseHear(LPCTSTR str)
Date    : 10.11.97
-------------------------------------------------------------------------------
Parameter: reference to the current position in the string which is 0
Returnval: TRUE, if no parse error occured
Description: parses a hear perception
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseHear(LPCTSTR str)
{
	TParseString	msg(str);
	CString			tmp;
	int				mode = 0;
	char*				strPlayMode[] = PLAYMODE_STRINGS;

	// get the time
	msg.SetPosition(5);
	m_iHearTime = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing heartime  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// get the source
	switch (msg.GetAt(msg.GetPosition()))
	{
		case 'r':
			// it is from the referee
			if (!msg.ParseString(tmp))
				return FALSE;
			// get the message
			if (!msg.ParseString(tmp))
				return FALSE;
			
			while ((mode < NO_OF_PLAYMODES) && (tmp.Compare(strPlayMode[mode])))
				mode++;

			if (mode < NO_OF_PLAYMODES)
			{
				// set the playmode
				m_iPlayMode = mode;
				if ((mode == CATCH_L) && (GetSide() == LEFT_TO_RIGHT))
					m_bGoalieCaught = TRUE;
				else if ((mode == CATCH_R) && (GetSide() == RIGHT_TO_LEFT))
					m_bGoalieCaught = TRUE;
				else if (mode == PLAY_ON)
					m_bGoalieCaught = FALSE;
			}
			else
			{
				// check if it is a goal, because goals have different text
				if (tmp.Left(6).Compare(strPlayMode[GOAL_L]) == 0)
					m_iPlayMode = GOAL_L;
				else if (tmp.Left(6).Compare(strPlayMode[GOAL_R]) == 0)
					m_iPlayMode = GOAL_R;
			}
			break;
		
		case 's':
			// it has been sent by myself
			break;
		
		default:
			// it is a say message
			if (!ParseSayMessage(&msg))
				return FALSE;
			break;
	}

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: ParseSayMessage(TParseString msg)
Date    : 27.05.1999
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: pointer on the message
Returnval: true if no error
Description: parses a say message from a teammate
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseSayMessage(TParseString *msg)
{
	int		iRecipient;
	int		iDispatcher;
	int		iStamina;
	CString	tmp;

	//overread direction
	msg->ParseInteger();

	//check if message from own team
	if (!msg->ParseString(tmp))
		return FALSE;
	if (tmp.Left(1) != '*')
		return TRUE;

	iRecipient = msg->ParseInteger();	//number of recipient 0 means for all
	if ( (iRecipient > 11) || (iRecipient < 0) )
		return FALSE;

	iDispatcher = msg->ParseInteger();	//dispatcher of the message
	if ( (iDispatcher > 11) || (iDispatcher < 1) )
		return FALSE;

	iStamina = msg->ParseInteger();		//dispatchers stamina
	m_pOwnTeam[iDispatcher-1]->SetStamina(iStamina);
	m_pOwnTeam[iDispatcher -1]->SetHearTimePlayer(m_iHearTime);

	//type of the message
	if (!msg->ParseString(tmp))
		return FALSE;
	switch (tmp.GetAt(0))
	{
		case 'P':		//say position message
//			if (m_pOwnTeam[iRecipient]->IsThisAgent())
				if (!ParseMsgPosition(msg))
					return FALSE;
		default:
		// it is a unknown message
			break;
	}

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: ParseMsgPosition(TParseString msg)
Date    : 27.05.1999
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: hearn message
Returnval: true in no error
Description: parses a position body
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseMsgPosition(TParseString *msg)
{
	int		i;
	int		length;
	int		iPlayerNr;
	double	dXpos;
	double	dYpos;
	
	//number of objects 2*11 players + 1 ball
	length = msg->ParseInteger();
	if ( (length > 23) || (length < 0) )
		return FALSE;

	for (i = 0; i < length; i++)
	{
		//gets the position and the object number
		//0 for ball
		//1 - 11 for own team
		//21 - 31 for other team
		iPlayerNr = msg->ParseInteger();
		dXpos = msg->ParseDouble();
		dYpos = msg->ParseDouble();

		if ( (iPlayerNr < 0) || (iPlayerNr > 31) 
			|| ((iPlayerNr > 11) && (iPlayerNr < 21)) )
			return FALSE;
		if (iPlayerNr == 0)			//ball
		{
			if (!m_pBall->IsValid() )
			{
				m_pBall->SetAbsX(dXpos);
				m_pBall->SetAbsY(dYpos);
				m_pBall->SetInMind();
				m_pBall->SetTime(m_iHearTime);
			}
		}
		else if (iPlayerNr < 20)	//players from own team
		{
			iPlayerNr -= 1;
			if (!m_pOwnTeam[iPlayerNr]->IsValid() )
			{
				m_pOwnTeam[iPlayerNr]->SetAbsX(dXpos);
				m_pOwnTeam[iPlayerNr]->SetAbsY(dYpos);
				m_pOwnTeam[iPlayerNr]->SetInMind();
				m_pOwnTeam[iPlayerNr]->SetTime(m_iHearTime);	
			}
		}
		else								//players from other team
		{
			iPlayerNr -= 21;
			if (!m_pOtherTeam[iPlayerNr]->IsValid() )
			{
				m_pOtherTeam[iPlayerNr]->SetAbsX(dXpos);
				m_pOtherTeam[iPlayerNr]->SetAbsY(dYpos);
				m_pOtherTeam[iPlayerNr]->SetInMind();
				m_pOtherTeam[iPlayerNr]->SetTime(m_iHearTime);
			}
		}
	}

	return TRUE;
}


/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseSenseBody(LPCTSTR str)
Date    : 27.01.98
-------------------------------------------------------------------------------
Parameter: string to parse
Returnval: FALSE, if an error occured
Description: parses a sense body information (sense_body TIME ((view_mode QUALITY
				 WIDTH) (stamina STAMINA EFFORT) (speed AMOUNT) (kick COUNT) (dash COUNT)
				 (turn COUNT) (say COUNT))
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseSenseBody(LPCTSTR str)
{
	TParseString	msg(str);
	CString			temp;
	int				time;


	// read the time
	msg.SetPosition(12);
	time = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing sensetime  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (time > m_iSenseTime)
		m_bTimeProceeding = TRUE;
	else
		m_bTimeProceeding = FALSE;
	m_iSenseTime = time;

	// read the view mode
	msg.ParseString(temp);
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing viewmode  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (temp.GetAt(0) == 'l')
		m_iViewQuality = 0;
	else
		m_iViewQuality = 1;
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing viewquality  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (temp.GetAt(1) == 'a')
		m_iViewAngle = 45;
	else if (temp.GetAt(0) == 'w')
		m_iViewAngle = 180;
	else
		m_iViewAngle = 90;

	// read the stamina
	msg.ParseString(temp);
	SetStamina((int) msg.ParseDouble());
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing stamina  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// read the effort
	m_dEffort = msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing effort  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_dEffort > 1.0)
		m_dEffort = 1.0;
	if (m_dEffort < EFFORT_MIN)
		m_dEffort = EFFORT_MIN;

	// read the speed
	msg.ParseString(temp);
	m_dSpeed = msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing speed  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_dSpeed > PLAYER_SPEED_MAX * PLAYER_DECAY)
		m_dSpeed = PLAYER_SPEED_MAX * PLAYER_DECAY;

	// read the headAngle
	msg.ParseString(temp);
	m_iHeadAngle = (int) msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing headangle  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// read the counts
	msg.ParseString(temp);
	m_iNoOfKick = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfKicks  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfKick < m_iExpectedNoOfKick)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("Kick not executed\n");
		#endif
	}
	m_iExpectedNoOfKick = m_iNoOfKick;
	
	msg.ParseString(temp);
	m_iNoOfDash = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfDash  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfDash < m_iExpectedNoOfDash)
	{
		// a dash has not been executed
		#ifdef _DEBUG
			TRACE0("Dash not executed\n");
		#endif
	}
	m_iExpectedNoOfDash = m_iNoOfDash;
	
	msg.ParseString(temp);
	m_iNoOfTurn = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfTurn  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfTurn < m_iExpectedNoOfTurn)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("Turn not executed\n");
		#endif
	}
	m_iExpectedNoOfTurn = m_iNoOfTurn;
	
	msg.ParseString(temp);
	m_iNoOfSay = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfSay  %s\r\n", msg);
		#endif
		return FALSE;
	}
	
	msg.ParseString(temp);
	m_iNoOfTurnNeck = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfTurnNeck  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfTurnNeck < m_iExpectedNoOfTurnNeck)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("TurnNeck not executed\n");
		#endif
	}
	m_iExpectedNoOfTurnNeck = m_iNoOfTurnNeck;

/*	char		buf1[200];
	char		buf2[200];
	double	dStamina;
	// (sense_body 3000 (view_mode high normal) (stamina 3500 1) (speed 0) (head_angle 0) (kick 0) (dash 0) (turn 7) (say 0) (turn_neck 0))
	int n = sscanf(str, "(sense_body %d (view_mode %s %s (stamina %f %f) (speed %f) (head_angle %d) (kick %d) (dash %d) (turn %d) (say %d) (turn_neck %d))"
						, &m_iSenseTime, buf1, buf2, &dStamina, &m_dEffort, &m_dSpeed, &m_iHeadAngle, &m_iNoOfKick, &m_iNoOfDash, &m_iNoOfTurnNeck);
*/	
	// are we waiting for this result (autorun)
	if (m_bWaitForLastSenseBody)
	{
		m_bWaitForLastSenseBody = FALSE;
		m_pAgent->OnReceivedLastSenseBody();
	}

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseInit(LPCTSTR str)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: string containing the init message (init side number playmode)
Returnval: FALSE, if an error occured
Description: parses the init message to get the side and players number
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseInit(LPCTSTR str)
{
	TParseString	msg(str);
	CString			tmp;
	int				number;
	int				mode;
	char*				strPlayMode[] = PLAYMODE_STRINGS;
	int				posLX[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_X;
	int				posLY[PLAYERS_PER_TEAM] = POS_KICKOFF_LEFT_Y;

	// get the side
	if ((msg.GetLength() > 6) && (msg.GetAt(6) == 'r'))
	{
		// the player is in the team playing from right to left
		SetSide(RIGHT_TO_LEFT);
		// swap x coordinates of own and other goal
		double tmp = m_pOtherGoal->GetAbsX();
		m_pOtherGoal->SetAbsX(m_pOwnGoal->GetAbsX());
		m_pOwnGoal->SetAbsX(tmp);
		// and of 11m points
		tmp = m_pOther11mPoint->GetAbsX();
		m_pOther11mPoint->SetAbsX(m_pOwn11mPoint->GetAbsX());
		m_pOwn11mPoint->SetAbsX(tmp);
	}
	else
		SetSide(LEFT_TO_RIGHT);

	// get the number of the player
	msg.SetPosition(8);
	number = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	ASSERT((number > 0) && (number <= PLAYERS_PER_TEAM));
	m_pAgent->SetNumber(number, m_iSide);
	m_pThisPlayer->SetTeam(TRUE, m_iSide);
	m_pThisPlayer->SetNum(number);
	// mark the player in our team with this number as this agent
	m_pOwnTeam[number - 1]->SetThisAgent(TRUE);

	// initialize the expected coordinates of own team' players
	for (int i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (m_iSide == LEFT_TO_RIGHT)
		{
			// it is a player of the left team
			m_pOwnTeam[i]->SetAbsX(posLX[i]);
			m_pOwnTeam[i]->SetAbsY(posLY[i]);
		}
		else
		{
			// it is a player of the right team, so mirror x coordinates
			m_pOwnTeam[i]->SetAbsX(-posLX[i]);
			m_pOwnTeam[i]->SetAbsY(-posLY[i]);
		}	
	}

	// get the playmode
	msg.ParseString(tmp);
	mode = 0;
	while ((mode < NO_OF_PLAYMODES) && (tmp.Compare(strPlayMode[mode])))
		mode++;

	if (mode < NO_OF_PLAYMODES)
		m_iPlayMode = mode;

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseObject(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: reference to the current position in the string, which must be the 
			  opening bracket of the objinfo
Returnval: TRUE, if no parse error occured
Description: parses an object and stores the information
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseObject(LPCTSTR str)
{
	TParseString	msg(str);
	TParseString	name;
	CString		teamname;
	BOOL			bMyTeam;
	int			team;
	int			iNum;
	double		dist;
	int			dir;
	double		distchng;
	double		dirchng;
	int			faceDir;
	int			bodyDir;
	int			index;
	int			type;
	char			tag1;
	char			tag2;
	int			flagNum;
	int			iFlag;
	TFlag*		pFlag;
	TLine*		pLine;

	// read the objects name bracket
	if (!msg.ParseBracket(name))
		return FALSE;
	
	// read the distance, direction and distchngection (the last may be not existing)
	if (m_iViewQuality == 0)
	{
		// we have got low view quality, so no distances
		dist = NOT_EXISTING;
		dir = msg.ParseInteger();
		if (msg.HasErrorOccured())
			return FALSE;
	}
	else
	{
		// we have got high view quality
		dist = msg.ParseDouble();
		if (msg.HasErrorOccured())
			return FALSE;
		dir = msg.ParseInteger();
		if (msg.HasErrorOccured())
			return FALSE;
	}
	dirchng = NOT_EXISTING;
	distchng = msg.ParseDouble();
	if (msg.HasErrorOccured())
		distchng = NOT_EXISTING;
	else 
	{	
		dirchng = msg.ParseDouble();
		if (msg.HasErrorOccured())
			dirchng = NOT_EXISTING;
		else 
		{	
			faceDir = msg.ParseInteger();
			if (msg.HasErrorOccured())
				faceDir = NOT_EXISTING;
			else 
			{	
				bodyDir = msg.ParseInteger();
				if (msg.HasErrorOccured())
					bodyDir = NOT_EXISTING;
			}
		}
	}

	// switch on the kind of object
	switch (name.GetAt(0))
	{
		// it is a player
		case 'P':
		case 'p':
			// read the name
			if (name.GetLength() > 7)
			{
				name.SetPosition(7);
				// read the teamname
				name.ParseString(teamname);
				// is it my team ?
				if (teamname.Compare(m_pAgent->GetTeam()) == 0)
					team = m_pAgent->GetSide();
				else if (m_pAgent->GetSide() == LEFT_TO_RIGHT)
					team = RIGHT_TO_LEFT;
				else
					team = LEFT_TO_RIGHT;

				// read player number
				iNum = name.ParseInteger();
				if (name.HasErrorOccured())
					iNum = NOT_EXISTING;
			}
			else
			{
				team = NOT_EXISTING;
				iNum = NOT_EXISTING;
			} 

			bMyTeam = (team == m_pThisPlayer->GetTeam());
			m_pVisiblePlayer[m_iNoOfPlayers]->SetPlayer(m_iSeeTime, dist, dir, distchng
													, dirchng, faceDir, bodyDir, bMyTeam, team, iNum);
			m_iNoOfPlayers++;

			break;


		case 'G':
		case 'g':
			// it is a goal
			if (name.GetLength() > 5)
			{
				// goal with name
				if (name.GetAt(5) == 'l')
					if (m_iSide == LEFT_TO_RIGHT)
						index = 0;
					else
						index = 1;
				else
					if (m_iSide == LEFT_TO_RIGHT)
						index = 1;
					else
						index = 0;
			}
			else
			{
				// it is a G goal (do not know which)
				if (m_pThisPlayer->GetAbsX() < 0.0)
				{
					if (m_iSide == LEFT_TO_RIGHT)
						index = 0;
					else
						index = 1;
				}
				else
				{
					if (m_iSide == LEFT_TO_RIGHT)
						index = 1;
					else
						index = 0;
				}
			}
			
			// store the values
			if (index == 0)
				m_pOwnGoal->SetVisibleObject(m_iSeeTime, dist, dir);
			else
				m_pOtherGoal->SetVisibleObject(m_iSeeTime, dist, dir);
						
			break;


		case 'B':
		case 'b':
			// it is the ball
			m_pBall->SetBall(m_iSeeTime, dist, dir, distchng, dirchng);
			break;

		case 'F':
			break;
		case 'f':
			// it is a flag
			type = FIELD_FLAG;
			if (name.GetLength() < 6)
				return FALSE;
			tag1 = name.GetAt(5);
			if (name.GetLength() < 8)
			{
				// it must be the center flag
				if (tag1 != 'c')
					return FALSE;
				iFlag = CENTER_MARK;
			}
			else
			{
				tag2 = name.GetAt(7);
				switch (tag1)
				{	
					case 'c':
						// it is a center flag 
						if (tag2 == 't')
							iFlag = CENTER_TOP;
						else
							iFlag = CENTER_BOTTOM;
						break;

					case 'g':
						// it is a goal flag
						tag1 = tag2;
						if (name.GetLength() < 10)
							return FALSE;
						tag2 = name.GetAt(9);
						if (tag1 == 'l')
						{
							if (tag2 == 't')
								iFlag = GOAL_POST_L;
							else
								iFlag = GOAL_POST_L + 1;
						}
						else
						{
							if (tag2 == 't')
								iFlag = GOAL_POST_R;
							else
								iFlag = GOAL_POST_R + 1;
						}
						break;

					case 'p':
						// it is a flag of the penalty area
						type = PENALTY_AREA_FLAG;
						tag1 = tag2;
						if (name.GetLength() < 10)
							return FALSE;
						tag2 = name.GetAt(9);
						if (tag1 == 'l')
						{
							if (tag2 == 't')
								iFlag = FLAG_PENALTY_L;
							else if (tag2 == 'c')
								iFlag = FLAG_PENALTY_L + 1;
							else
								iFlag = FLAG_PENALTY_L + 2;
						}
						else
						{
							if (tag2 == 't')
								iFlag = FLAG_PENALTY_R;
							else if (tag2 == 'c')
								iFlag = FLAG_PENALTY_R + 1;
							else
								iFlag = FLAG_PENALTY_R + 2;
						}
						
						break;

					case 'l':
						// it is one of the left flags of the field
						if (tag2 == '0')
							// it is the left center flag
							iFlag = FLAG_OUTSIDE_LC;
						else
						{
							name.SetPosition(9);
							flagNum = name.ParseInteger();
							if (name.HasErrorOccured())
								// it is an old lt or lb flag
								return TRUE;
							flagNum /= 10;
							flagNum--;
							if ((flagNum < 0) || flagNum > 2)
								return FALSE;
							if (tag2 == 't')
								// it is a left top outside flag
								iFlag = FLAG_OUTSIDE_LT + flagNum;
							else
								// it is a left bottom outside flag
								iFlag = FLAG_OUTSIDE_LB + flagNum;
						}
						break;

					case 'r':
						// it is one of the right flags of the field
						if (tag2 == '0')
							// it is the right center flag
							iFlag = FLAG_OUTSIDE_RC;
						else
						{
							name.SetPosition(9);
							flagNum = name.ParseInteger();
							if (name.HasErrorOccured())
								// it is an old rt or rb flag
								return TRUE;
							flagNum /= 10;
							flagNum--;
							if ((flagNum < 0) || flagNum > 2)
								return FALSE;
							if (tag2 == 't')
								// it is a right top outside flag
								iFlag = FLAG_OUTSIDE_RT + flagNum;
							else
								// it is a right bottom outside flag
								iFlag = FLAG_OUTSIDE_RB + flagNum;
						}
						break;

					case 't':
						// it is one of the top flags of the field
						if (tag2 == '0')
							// it is the top center flag
							iFlag = FLAG_OUTSIDE_TC;
						else
						{
							name.SetPosition(9);
							flagNum = name.ParseInteger();
							if (name.HasErrorOccured())
								return FALSE;
							flagNum /= 10;
							flagNum--;
							if ((flagNum < 0) || flagNum > 4)
								return FALSE;
							if (tag2 == 'l')
								// it is a top left outside flag
								iFlag = FLAG_OUTSIDE_TL + flagNum;
							else
								// it is a top right outside flag
								iFlag = FLAG_OUTSIDE_TR + flagNum;
						}
						break;

					case 'b':
						// it is one of the bottom flags of the field
						if (tag2 == '0')
							// it is the bottom center flag
							iFlag = FLAG_OUTSIDE_BC;
						else
						{
							name.SetPosition(9);
							flagNum = name.ParseInteger();
							if (name.HasErrorOccured())
								return FALSE;
							flagNum /= 10;
							flagNum--;
							if ((flagNum < 0) || flagNum > 4)
								return FALSE;
							if (tag2 == 'l')
								// it is a bottom left outside flag
								iFlag = FLAG_OUTSIDE_BL + flagNum;
							else
								// it is a bottom right outside flag
								iFlag = FLAG_OUTSIDE_BR + flagNum;
						}
						break;

					default:
						iFlag = TRASH_FLAG;
				}
			}

			pFlag = m_pFlag[iFlag];
			pFlag->SetVisibleObject(m_iSeeTime, dist, dir);
			pFlag->SetFlag(type, tag1, tag2);

			break;

		case 'l':
			// it is a line
			// we do not use lines up to now
			pLine = NULL;
			/*			if (name.GetLength() < 6)
				break;
			tag1 = name.GetAt(5);
			switch (tag1)
			{
				case 'l':
					pLine = m_pLine[0];
					break;

				case 'r':
					pLine = m_pLine[1];
					break;

				case 't':
					pLine = m_pLine[2];
					break;

				case 'b':
					pLine = m_pLine[3];
					break;
			}

			pLine->SetVisibleObject(m_iSeeTime, dist, dir);
*/			break;
	}

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::Localize()
Date    : 16.02.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the absolute position of the agent could be calculated
Description: calculates the absolute position of the agent. Two flags are 
				 necessary (with this algorithm) to locate the position. It is at one
				 of the two intersection points of the circles surounding the flags
				 in the distance of the player. Which intersection is decided by the 
				 relative angle the player sees the flags.
				 To calculate the intersection points, first the inbtersection point P3 of 
				 the line from flag1 to flag2 and the line from intersection point 1
				 and 2 is calculated. 
				 
-----------------------------------------------------------------------------*/
BOOL TPerception::Localize()
{
	int		i;
	int		iFlag1 = -1;
	int		iFlag2 = -1;
	int		iMaxAngle = -1000;
	int		iDeltaAngle;
	int		iFlag1Direction;
	double	r1, r2;				// the distance of the player to the flags 
	double	dist, dist2;		// the distance (square) of the two flags
	double	a;						// distance from one flag to intersection point P3
	double	h;						// distance from P3 to the intersection Points P1 and P2 of the two circles
	double	x, x1, x2, x3;
	double	y, y1, y2, y3;
	double	ratio;
	int		iAbsDirection;
	int		beta;

	// search for two visible flags with maximal difference of the view angle
	for (i = 0; i < NO_OF_FLAGS; i++)
	{
		if (m_pFlag[i]->IsValid() && (m_pFlag[i]->GetDistance() > 4.0))
		{
			if (iFlag1 == -1)
			{
				// found the first visible flag
				iFlag1 = i;
				iFlag1Direction = m_pFlag[iFlag1]->GetDirection();
			}
			else
			{
				iDeltaAngle = iFlag1Direction - m_pFlag[i]->GetDirection();
				if (iDeltaAngle < 0)
					iDeltaAngle = -iDeltaAngle;
				if (iDeltaAngle > iMaxAngle)
				{
					// this flag has a bigger angle to the first flag
					iFlag2 = i;
					iMaxAngle = iDeltaAngle;
				}
			}
		}
	}

	if (iFlag2 == -1)
		// there are not 2 visible flags
		return FALSE;

	// do the calculations
	r1 = m_pFlag[iFlag1]->GetDistance();
	r2 = m_pFlag[iFlag2]->GetDistance();

	x1 = m_pFlag[iFlag1]->GetAbsX();
	x2 = m_pFlag[iFlag2]->GetAbsX();

	y1 = m_pFlag[iFlag1]->GetAbsY();
	y2 = m_pFlag[iFlag2]->GetAbsY();

	// calculate the square distance of the two flags
	dist2 = (x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1);	
	dist = sqrt(dist2);
	if (dist > r1 + r2)
	{
		// the circles would not intersect
		dist = r1 + r2;
		dist2 = dist * dist;
	}
	else if ((r1 > r2) && (dist + r2 < r1))
	{
		// the circles would not intersect
		dist = r1 - r2;
		dist2 = dist * dist;
	}
	else if ((r2 > r1) && (dist + r1 < r2))
	{
		// the circles would not intersect
		dist = r2 - r1;
		dist2 = dist * dist;
	}

	r1 *= r1;
	r2 *= r2;
	
	// a = (r1^2 - r2^2 + d^2 ) / (2 d) 
	a = (r1 - r2 + dist2) / (2.0 * dist);
	// h^2 = r1^2 - a^2
	h = r1 - a*a;
	if (h < 0.0)
		h = 0.0;
	else
		h = sqrt(h);

	// P3 = P1 + a ( P2 - P1 ) / d
	x3 = x1 + a * (x2 - x1) / dist;
	y3 = y1 + a * (y2 - y1) / dist;

	// two circles intersect usually in 2 points. Find out which one to select
	if (m_pFlag[iFlag1]->GetDirection() > m_pFlag[iFlag2]->GetDirection())
	{
		// result x = x3 + h ( y2 - y1 ) / d 
		x = x3 + h * (y2 - y1) / dist;
		// result y = y3 - h ( x2 - x1 ) / d 
		y = y3 - h * (x2 - x1) / dist;
	}
	else
	{
		x = x3 - h * (y2 - y1) / dist;
		y = y3 + h * (x2 - x1) / dist;
	}

	// calculate the absolute direction
	r1 = m_pFlag[iFlag1]->GetDistance();
	ratio = (y1 - y) / r1;
	beta = (int) (asin(ratio) * RAD2DEG);
	if (x > x1)
		iAbsDirection = 180 - beta - iFlag1Direction;
	else
		iAbsDirection = beta - m_pFlag[iFlag1]->GetDirection();

	// set this players coordinates
	m_pThisPlayer->SetPlayerAbsCoordinates(TVector(x, y), iAbsDirection, m_iHeadAngle, 
							m_dSpeed, m_iViewAngle,	m_pAgent->IsDashingBackward());

	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::GlobalMap()
Date    : 18.05.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates the absolute position of all seen moveable objects
-----------------------------------------------------------------------------*/
void TPerception::GlobalMap()
{
	int	i;
	int	dir = m_pThisPlayer->GetAbsDir();
	
	// set balls global coordinates
	if (m_pBall->IsValid())
		m_pBall->SetAbsCoordinates(m_pThisPlayer, dir);

	// calculate the new future positions of the ball
	for (i = 0; i < MAX_FUTURE_POSITIONS; i++)
		m_pBall->GetFuturePosition(i, *m_pFutureBallPosition[i]);
	
	for (i = 0; i < m_iNoOfPlayers; i++)
		if (m_pVisiblePlayer[i]->IsValid())
			m_pVisiblePlayer[i]->SetAbsCoordinates(m_pThisPlayer, dir);	
}


/*-----------------------------------------------------------------------------
Function: void TPerception::CheckVisualPlausibility()
Date    : 03.06.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: removes players from the mind map, that should be in the visible area, but
				 are not seen
-----------------------------------------------------------------------------*/
void TPerception::CheckVisualPlausibility()
{
	int	i;
	int	iMaxAngle = m_iViewAngle / 2 - 10;

	// check if the player or ball should be visible, but is not
	m_pThisPlayer->CheckVisuality(m_pBall, iMaxAngle, m_iSeeTime);

	// switched off 27.7.98, because players too far to see teamname are treted incorrect
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (!m_pOwnTeam[i]->IsThisAgent())
			m_pThisPlayer->CheckVisuality(m_pOwnTeam[i], iMaxAngle, m_iSeeTime);
		m_pThisPlayer->CheckVisuality(m_pOtherTeam[i], iMaxAngle, m_iSeeTime);
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::GetIsAt(int x, int y, int dist1) const
Date    : 17.03.98
-------------------------------------------------------------------------------
Parameter: coordinates to check against
			  error acceptable
Returnval: TRUE, if the agent is, within some boundary, at x, y
Description: coordinates are compared with the coordinates, the agent beliefes
				 to be in. if the difference is less than dist, True is returned
-----------------------------------------------------------------------------*/
BOOL TPerception::GetIsAt(int x, int y, int dist1) const
{
	double	dx = (double) x - m_pThisPlayer->GetAbsX();
	double	dy;
	double	dist;

	// check x coordinate
	dist = (double) dist1;
	if ((dx < -dist) || (dx > dist))
		return FALSE;

	// check y coordinate
	dy = (double) y - m_pThisPlayer->GetAbsY();
	if ((dy < -dist) || (dy > dist))
		return FALSE;
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::MatchPlayers()
Date    : 12.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: tries to match the perceived players to players remembered
				 May be this is not very important in eleven players matches, because
				 there are enough important players fully visible
-----------------------------------------------------------------------------*/
void TPerception::MatchPlayers()
{
	int	i;
	int	team;
	int	iNum;

	// invalidate match info
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i]->SetMatched(FALSE);
		m_pOtherTeam[i]->SetMatched(FALSE);
	}

	// treat all the players thats numbers are seen
	for (i = 0; i < m_iNoOfPlayers; i++)
	{
		team = m_pVisiblePlayer[i]->GetTeam();
		iNum = m_pVisiblePlayer[i]->GetNum();

		// try to identify the player
		if (iNum != NOT_EXISTING)
		{
			// team and number can be seen
			ASSERT((iNum > 0) && (iNum <= PLAYERS_PER_TEAM));
			if (team == m_iSide)
				m_pOwnTeam[iNum - 1]->SetPlayer(m_pVisiblePlayer[i]);
			else
				m_pOtherTeam[iNum - 1]->SetPlayer(m_pVisiblePlayer[i]);
		}
	}

	// treat all the players thats numbers are not seen
	for (i = 0; i < m_iNoOfPlayers; i++)
	{
		team = m_pVisiblePlayer[i]->GetTeam();
		iNum = m_pVisiblePlayer[i]->GetNum();

		// try to identify the player
		if (iNum == NOT_EXISTING)
			// find the remembered player closest to the viewed player
			MatchClosestPlayer(m_pVisiblePlayer[i], team);
	}
}

/*-----------------------------------------------------------------------------
Function: void TPerception::MatchClosestPlayer(TPlayer* pViewedPlayer, int team)
Date    : 12.03.98
-------------------------------------------------------------------------------
Parameter: pointer to the player that has been perceived
			  team index: 0 if own team, 1 if other team, NOT_EXISTING if unknown
Returnval: (void)
Description: tries to find the remembered player closest to the perceived player
-----------------------------------------------------------------------------*/
void TPerception::MatchClosestPlayer(TPlayer* pViewedPlayer, int team)
{
	int		i;
	int		iNearest = 0;
	double	dNearestDist = 10000.0;
	double	dDist;
	int		iFreeOpponent;

	if (team == m_iSide) 
	{
		// it is a player of our team
		for (i = 0; i < PLAYERS_PER_TEAM; i++)
		{
			if (!m_pOwnTeam[i]->IsThisAgent() && !m_pOwnTeam[i]->IsMatched())
			{
				// it is not me self and not a player already matched
				dDist = pViewedPlayer->GetDistanceTo(m_pOwnTeam[i]);
				if (dDist < dNearestDist)
				{
					iNearest = i;
					dNearestDist = dDist;
				}
			}
		}
	}
	else if (team != NOT_EXISTING) 
	{
		// it is a player of the opponent team
		for (i = 0; i < PLAYERS_PER_TEAM; i++)
		{
			if (!m_pOtherTeam[i]->IsMatched())
			{
				dDist = pViewedPlayer->GetDistanceTo(m_pOtherTeam[i]);
				if (dDist < dNearestDist)
				{
					iNearest = i;
					dNearestDist = dDist;
				}
			}
		}
	}

	
	else 
		// player does not know which team
		//	switched off 22.07.98. In eleven player teams, one can rely on visible players
		return;
	
	// the player's parameters have to be set	
	if (team == m_iSide)
	{
		m_pOwnTeam[iNearest]->SetPlayer(pViewedPlayer);
		m_pOwnTeam[iNearest]->SetTeam(TRUE, team);
		m_pOwnTeam[iNearest]->SetNum(iNearest + 1);
		m_pOwnTeam[iNearest]->SetOnlyInMind();
	}
	else
	{
		if (m_pOtherTeam[iNearest]->IsValid())
		{
			// the player iNearest has been matched previously (errorneously)
			iFreeOpponent = GetFreeOtherTeam();
			if (iNearest < PLAYERS_PER_TEAM)
			{
				m_pOtherTeam[iFreeOpponent]->SetPlayer(m_pOtherTeam[iNearest]);
				m_pOtherTeam[iFreeOpponent]->SetNum(iFreeOpponent + 1);
			}
		}
		m_pOtherTeam[iNearest]->SetPlayer(pViewedPlayer);
		m_pOtherTeam[iNearest]->SetTeam(FALSE, team);
		m_pOtherTeam[iNearest]->SetNum(iNearest + 1);
		m_pOtherTeam[iNearest]->SetOnlyInMind();
	}
}

/*-----------------------------------------------------------------------------
Function: int TPerception::GetFreeOtherTeam()
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: number of a m_pOtherTeam opponent, that has not been matched (is not valid)
Description: tries to find an opponent, that has not been matched this perception.
				 All players of unknown team are opponents !
-----------------------------------------------------------------------------*/
int TPerception::GetFreeOtherTeam()
{
	int	i = 0;

	while ((i < PLAYERS_PER_TEAM) && (m_pOtherTeam[i]->IsValid()))
		i++;

	return i;
}

/*-----------------------------------------------------------------------------
Function: int TPerception::GetPlayersWithin(TDirection& dir1, TDirection& dir2)
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: left direction of sector
			  right border of sector
Returnval: number of players within the specified sector
Description: counts the players within the specified sector
-----------------------------------------------------------------------------*/
int TPerception::GetPlayersWithin(TDirection dir1, TDirection dir2)
{
	int			i;
	TDirection	dir;
	int			count = 0;

	if (dir1 > dir2)
	{
		int temp = dir2;
		dir2 = dir1;
		dir1 = temp;
	}

	// first count the opponents
	count = GetOpponentsWithin(dir1, dir2);

	// now count own team
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (m_pOwnTeam[i]->IsInMind())
		{
			dir = GetObjectDirection(m_pOwnTeam[i]);
			if ((dir >= dir1) && (dir <= dir2))
				count++;
		}
	}

	return count;
}

/*-----------------------------------------------------------------------------
Function: int TPerception::GetOpponentsWithin(TDirection dir1, TDirection dir2)
Date    : 31.08.98
-------------------------------------------------------------------------------
Parameter: left direction of slice
			  right direction of slice
Returnval: (void)
Description: counts the numbers of opponent players, that lie within the two directions
-----------------------------------------------------------------------------*/
int TPerception::GetOpponentsWithin(TDirection dir1, TDirection dir2)
{
	int			i;
	TDirection	dir;
	int			count = 0;

	if (dir1 > dir2)
	{
		int temp = dir2;
		dir2 = dir1;
		dir1 = temp;
	}

	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (m_pOtherTeam[i]->IsInMind())
		{
			dir = GetObjectDirection(m_pOtherTeam[i]);
			if ((dir >= dir1) && (dir <= dir2))
				count++;
		}
	}
	return count;
}

/*-----------------------------------------------------------------------------
Function: Functions returning properties of the world
Date    : 24.02.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: property
Description: 
-----------------------------------------------------------------------------*/
TDirection TPerception::GetObjectDirection(TVisibleObject* obj)
{
	// 06.07.99 allways take global map
	// if the player can see the object, then return visible direction
//	if (obj->IsValid() && (m_iSenseTime - m_iSeeTime == 0))
//		return obj->GetDirection(); 
//	else
		// the object is not visible, so use the absolute position of the agent
		return m_pThisPlayer->GetAngleTo(obj);
}

TDirection TPerception::GetObjectDirectionToBody(TVisibleObject* obj)
{
	// 06.07.99 allways take global map
	// returns the objects direction with respect to the body direction of the agent
//	if (obj->IsValid() && (m_iSenseTime - m_iSeeTime == 0))
//		return obj->GetDirection() + m_iHeadAngle;
//	else
		// the object is not visible, so use the absolute position of the agent
		return m_pThisPlayer->GetAngleTo(obj) + m_iHeadAngle;
}

double TPerception::GetObjectDistance(TVisibleObject* obj)
{ 
	if (obj->IsValid() && (m_iSenseTime - m_iSeeTime == 0))
		return obj->GetDistance(); 
	else
		return m_pThisPlayer->GetDistanceTo(obj);
}

TDirection TPerception::GetObjectDeltaDir(TVisibleObject* obj)
{
	return obj->GetDeltaDir(); 
}

double TPerception::GetObjectDeltaDist(TVisibleObject* obj)
{ 
	return obj->GetDeltaDist(); 
}

BOOL TPerception::GetSeeObject(TVisibleObject* obj) const
{
	if (m_iSeeTime == obj->GetTime()) 
		if ((m_iSeeTime == 0) && (obj->GetAbsX() < -500.0))
			return FALSE;
		else
			return TRUE; 
	else 
		return FALSE; 
}

TSoccerGoal* TPerception::GetOwnGoal()
{	
	return m_pOwnGoal;
}

TSoccerGoal* TPerception::GetOtherGoal()
{	
	return m_pOtherGoal;
}

TFlag* TPerception::GetOtherLeftCorner()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[FLAG_OUTSIDE_TR + 4];
	else
		return m_pFlag[FLAG_OUTSIDE_BL + 4]; 
}

TFlag* TPerception::GetOtherRightCorner()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[FLAG_OUTSIDE_BR + 4];
	else
		return m_pFlag[FLAG_OUTSIDE_TL + 4]; 
}

TFlag* TPerception::GetOtherLeftPenaltyCorner()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[FLAG_PENALTY_R];
	else
		return m_pFlag[FLAG_PENALTY_L + 2]; 
}

TFlag* TPerception::GetOtherRightPenaltyCorner()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[FLAG_PENALTY_R + 2];
	else
		return m_pFlag[FLAG_PENALTY_L]; 
}

TFlag* TPerception::GetOtherLeftGoalPost()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[GOAL_POST_R];
	else
		return m_pFlag[GOAL_POST_L + 1]; 
}

TFlag* TPerception::GetOtherRightGoalPost()
{	
	if (m_pAgent->GetSide() == LEFT_TO_RIGHT) 
		return m_pFlag[GOAL_POST_R + 1];
	else
		return m_pFlag[GOAL_POST_L]; 
}

int TPerception::GetBodyAngle() const
{
	// returns the absolute body direction
	TDirection dir = m_pThisPlayer->GetAbsDir() - m_iHeadAngle;
	return dir;
}

/*-----------------------------------------------------------------------------
Function: TPlayer* TPerception::GetPlayerInWay(int dir)
Date    : 13.03.98
-------------------------------------------------------------------------------
Parameter: direction which is relevant
Returnval: pointer to the player, that is in the direction passed and near this 
			  agent, NULL, if none such exists
Description: searches for a player, that is within a certain direction of this player
				 and is near this player. It is used to avoid collisions
-----------------------------------------------------------------------------*/
TPlayer* TPerception::GetPlayerInWay(int dir)
{
	int		iMinDir = 1000;
	double	dMinDist;
	double	dist;
	int		iIndexMinDir = -1;
	int		iTeam = 0;
	int		i;
	int		deltaDir;


	// check the own team
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if ((m_pOwnTeam[i]->GetAge(m_iSeeTime) < 10) && !m_pOwnTeam[i]->IsThisAgent())
		{
			dist = GetObjectDistance(m_pOwnTeam[i]);
			if (dist < 5.0)
			{
				// find the player closest to the passed direction, that is within
				// five meters of this player
				deltaDir = dir - GetObjectDirection(m_pOwnTeam[i]);
				if (deltaDir < 0)
					deltaDir = -deltaDir;
				if (deltaDir < iMinDir)
				{
					iMinDir = deltaDir;
					dMinDist = dist;
					iIndexMinDir = i;
					iTeam = 0;
				}
			}
		}
	}

	// check the other team
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (m_pOtherTeam[i]->GetAge(m_iSeeTime) < 10)
		{
			dist = GetObjectDistance(m_pOtherTeam[i]);
			if (dist < 5.0)
			{
				// find the player closest to the passed direction, that is within
				// five meters of this player
				deltaDir = dir - GetObjectDirection(m_pOtherTeam[i]);
				if (deltaDir < 0)
					deltaDir = -deltaDir;
				if (deltaDir < iMinDir)
				{
					iMinDir = deltaDir;
					dMinDist = dist;
					iIndexMinDir = i;
					iTeam = 1;
				}
			}
		}
	}

	if ((iMinDir < 90) && (dMinDist * sin((double)iMinDir * DEG2RAD) < 3.0))
		if (iTeam == 0)
			return m_pOwnTeam[iIndexMinDir];
		else
			return m_pOtherTeam[iIndexMinDir];
	else
		return NULL;
}

/*-----------------------------------------------------------------------------
Function: TPlayer* TPerception::CalculateFreestTeammate(int dir)
Date    : 22.06.98
From    :
Modified: MP 24.04.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: pointer to the player, that has highest degree of freeness
			  null, if there is no teammate, or the highest degree of freeness is too low
Description: returns a pointer to the teammate, that has got the highest degree
				 of freeness or null, if it is too low
-----------------------------------------------------------------------------*/
TPlayer* TPerception::CalculateFreestTeammate()
{
	int		i;
	double	value;
	double	dHighest = 0.0;
	int		iHighest = -1;

	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if ((m_pOwnTeam[i]->IsInMind()) && !m_pOwnTeam[i]->IsThisAgent()
			&& (m_pOwnTeam[i]->IsAlife(m_iSenseTime)))

		value = CalculateFreePlayer(m_pOwnTeam[i]);
		if (value > dHighest)
		{
			dHighest = value;
			iHighest = i;
		}
	}

	// check for threshold
	if (iHighest < 0)
		return NULL;
	else
		return m_pOwnTeam[iHighest];
}

/*-----------------------------------------------------------------------------
Function: double	TPerception::CalculateFreePlayer (TPlayer * pTeammate)
Date    : 24.04.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: for which teammate is the calculation
Returnval: quality of freeness of this player
Description: 
-----------------------------------------------------------------------------*/
double	TPerception::CalculateFreePlayer (TPlayer * pTeammate)
{
	int i;
	int iSecureBoundary = RANGE_DEGREE_ACCESS_BALL;
	double dQuality;
	double dQualityOfPossibility;		// 0.5
	double dQualityOfPassDir;	// 0.2
	double dQualityOfMateDistance;	// 0.1
	double dQualityOfGoalDistance;
	double dDistanceToOwnGoal;

	double dDist;
	TDirection iWidth;
	TVector	tOptimalPassPoint;
	int		iDeltaStart;
	int		iDeltaStop;

	//calculate quality of possibility
	if (pTeammate->GetDistance() < MAX_SHOT_RANGE)
	{
		pTeammate->InitFreePassArea (iSecureBoundary);
		for (i = 0; i < PLAYERS_PER_TEAM; i++)
		{
			if ((m_pOtherTeam[i]->IsInMind())
				&&(this->GetObjectDistance(m_pOtherTeam[i]) < this->GetObjectDistance(pTeammate)))
			{
				pTeammate->UpdateFreePassArea(this->GetObjectDirection(m_pOtherTeam[i]), iSecureBoundary);
			}
		}
		dQualityOfPossibility = (double)pTeammate->GetFreePassArea() / ((double)iSecureBoundary*2);
	}
	else
		dQualityOfPossibility = 0.0;

	//set quality of distance to own goal
	dDistanceToOwnGoal = pTeammate->GetDistanceTo(m_pOwnGoal);
	dQualityOfGoalDistance = m_pAgent->GetLinearFuzzyValue(dDistanceToOwnGoal, 20.0, 25.0);

	//calculate quality of direction to the other goal
	iDeltaStart = 50;
	iDeltaStop = 80;
	GetOptimalPassPoint(tOptimalPassPoint, iDeltaStart, iDeltaStop);
	iWidth = this->GetObjectDirection(pTeammate) - m_pThisPlayer->GetAngleTo(tOptimalPassPoint);
	iWidth = iWidth.Abs();
	dQualityOfPassDir = m_pAgent->GetLinearFuzzyValue(iWidth, iDeltaStart, iDeltaStop, FALSE);

	//calculate quality of distance to the teammate
	dDist = (int)this->GetObjectDistance(pTeammate);
	if (dDist < TMATE_MAX_DIST)
		dQualityOfMateDistance = m_pAgent->GetLinearFuzzyValue(dDist, TMATE_MIN_DIST, TMATE_MID_DIST);
	else
		dQualityOfMateDistance = m_pAgent->GetLinearFuzzyValue(dDist, TMATE_MAX_DIST, MAX_SHOT_RANGE, FALSE);
	
	//calculation of the quality
	dQuality = dQualityOfPossibility * dQualityOfPassDir * dQualityOfMateDistance
					* dQualityOfGoalDistance;

	dQuality = pow(dQuality, 0.25);

	pTeammate->SetFreePassDegree(dQuality);

	return dQuality;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::GetOptimalPassPoint(TVector& point)
Date    : 08.07.99
-------------------------------------------------------------------------------
Parameter: reference to the vector that gets the result
			  reference to the max delta direction that has value 1.0
			  reference to the min delta direction that has value 0.0
Returnval: (void)
Description: calculates the optimal point the direction to which is optimal to pass
-----------------------------------------------------------------------------*/
void TPerception::GetOptimalPassPoint(TVector& point, int& iDeltaStart, int& iDeltaStop)
{
	int	x = (int) m_pThisPlayer->GetAbsX();
	int	y = (int) m_pThisPlayer->GetAbsY();

	if (GetSide() == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	if (x < -30)
	{
		// near the own goal, the player should pass to one of the midline posts
		if (y > 0)
			point = GetOtherRightCorner()->GetPosition();
		else
			point = GetOtherLeftCorner()->GetPosition();

		iDeltaStart = 25;
		iDeltaStop = 45;
	}
	else if (x < -15)
	{
		// near the own goal, the player should pass to one of the midline posts
		if (y > 0)
			point = GetOtherRightPenaltyCorner()->GetPosition();
		else
			point = GetOtherLeftPenaltyCorner()->GetPosition();

		iDeltaStart = 35;
		iDeltaStop = 60;
	}
	else if ((x > 30) && ((y < -10) || (y > 10)))
	{
		// we are close to the goal, pass to 11m point
		point = GetOther11mPoint()->GetPosition();
		iDeltaStart = 20;
		iDeltaStop = 40;
	}
	else
	{
		// in the rest of the own half, the pass direction is the other goal
		point = GetOtherGoal()->GetPosition();
		iDeltaStart = 40;
		iDeltaStop = 90;
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::IsPlayerTired(TPlayer* player)
Date    : 12.07.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: true, if the player can not run to the ball, because it is tired
Description: looks if the player is fit enough to run to the ball 
-----------------------------------------------------------------------------*/
BOOL TPerception::IsPlayerTired(TPlayer* player)
{
	double	ballDist = player->GetDistanceTo(m_pBall);
	int		stamina = player->GetStamina();

	if (stamina - ballDist * 85 < LOW_STAMINA - 300)
		return TRUE;
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::CalculateFunctionalObjects()
Date    : 13.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: void
Description: calculates, which object has got which function in the environment. 
				 E.g. the opponent at the ball or the teammate nearest to me
-----------------------------------------------------------------------------*/
void TPerception::CalculateFunctionalObjects()
{
	// check the own team
	CheckOwnTeamForFunctionalObjects();

	// decay offside
	CheckBallForOffside();

	// check the other team
	CheckOtherTeamForFunctionalObjects();
}

/*-----------------------------------------------------------------------------
Function: void TPerception::CheckOwnTeamForFunctionalObjects()
Date    : 14.04.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates functional objects of agent's own team
-----------------------------------------------------------------------------*/
void TPerception::CheckOwnTeamForFunctionalObjects()
{
	int		i;
	int		iDistBallTeammate;
	int		iMinDistBallTeammate = 1000;
//	int		iDistMeTeammate;
//	int		iMinDistMeTeammate = 1000;
	int		x = (int) m_pThisPlayer->GetAbsX();
	int		y = (int) m_pThisPlayer->GetAbsY();
	double	dTeammateX;
	double	dFrontmostX = 0.0;
	double	dBackmostX = 0.0;
	double	d2BackmostX = 0.0;
	
	// make coordinates team independent
	if (m_iSide == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	m_pTeammateAtBall = NULL;
	m_pTeammateAtMe = NULL;
	m_pFrontmostTeammate = NULL;
	m_pBackmostTeammate = NULL;
	m_p2BackmostTeammate = NULL;

	// check the own team
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if ((m_pOwnTeam[i]->IsInMind()) && !m_pOwnTeam[i]->IsThisAgent())
		{
			// we have a teammate that is in mind and is not this agent itself
			// look for the teammate at the ball
			if (m_pOwnTeam[i]->IsAlife(m_iSenseTime) && !IsPlayerTired(m_pOwnTeam[i]))
			{
				iDistBallTeammate = (int) m_pOwnTeam[i]->GetDistanceTo(m_pBall);
				if (iDistBallTeammate < iMinDistBallTeammate)
				{
					iMinDistBallTeammate = iDistBallTeammate;
					m_pTeammateAtBall = m_pOwnTeam[i];
				}
			}

			// look for frontmost player
			if (m_iSide == RIGHT_TO_LEFT)
				dTeammateX = m_pOwnTeam[i]->GetAbsX();
			else
				dTeammateX = -m_pOwnTeam[i]->GetAbsX();
			if (dTeammateX < dFrontmostX)
			{
				dFrontmostX = dTeammateX;
				m_pFrontmostTeammate = m_pOwnTeam[i];
			}

			// look for backmost players
			if (m_iSide == RIGHT_TO_LEFT)
				dTeammateX = m_pOwnTeam[i]->GetAbsX();
			else
				dTeammateX = -m_pOwnTeam[i]->GetAbsX();
			if (dTeammateX > d2BackmostX)
			{
				if (dTeammateX > dBackmostX)
				{
					// backmost player
					d2BackmostX = dBackmostX;
					dBackmostX = dTeammateX;
					m_p2BackmostTeammate = m_pBackmostTeammate;
					m_pBackmostTeammate = m_pOwnTeam[i];
				}
				else
				{
					d2BackmostX = dTeammateX;
					m_p2BackmostTeammate = m_pOwnTeam[i];
				}
			}
		}
	}
	m_pFreeTeammate = CalculateFreestTeammate();
}

/*-----------------------------------------------------------------------------
Function: void TPerception::CheckOtherTeamForFunctionalObjects()
Date    : 14.04.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: checks the other team for functional objects
-----------------------------------------------------------------------------*/
void TPerception::CheckOtherTeamForFunctionalObjects()
{
	int		i;
//	int		j;
	double	dDistBallOpponent;
	double	dMinDistBallOpponent = 1000.0;
	double	dDistMeOpponent;
	double	dMinDistMeOpponent = 1000.0;
	int		iNoOfOffsidePreventingPlayers;
	double	dOpponentX;
	double	dBackmostX = 0.0;
	double	d2BackmostX = 0.0;
	int		x = (int) m_pThisPlayer->GetAbsX();
	int		y = (int) m_pThisPlayer->GetAbsY();

	// make coordinates team independent
	if (m_iSide == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	iNoOfOffsidePreventingPlayers = 0;
	m_pOpponentAtBall = NULL;
	m_pOpponentAtMe = NULL;
	m_pBackmostOpponent = NULL;
	m_p2BackmostOpponent = NULL;

		// check the other team
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		if (m_pOtherTeam[i]->IsInMind())
		{
			// look for the opponent at the ball
			dDistBallOpponent = m_pOtherTeam[i]->GetDistanceTo(m_pBall);
			if (dDistBallOpponent < dMinDistBallOpponent)
			{
				dMinDistBallOpponent = dDistBallOpponent;
				m_pOpponentAtBall = m_pOtherTeam[i];
			}

			// look for Opponent at me
			dDistMeOpponent = GetObjectDistance(m_pOtherTeam[i]);
			if (dDistMeOpponent < dMinDistMeOpponent)
			{
				dMinDistMeOpponent = dDistMeOpponent;
				m_pOpponentAtMe = m_pOtherTeam[i];
			}

			// check if this player prevents me from beeing offside
			if (!m_bKnowNotInOffside && m_pThisPlayer->CheckPreventOffside(m_pOtherTeam[i]))
			{
				// there have to be two players behind this agent (goalie + another)
				iNoOfOffsidePreventingPlayers++;
				if (iNoOfOffsidePreventingPlayers > 1)
				{
					m_bKnowNotInOffside = TRUE;
					m_dInOffside = 0.0;
					m_iCheckOffsideCounter = 0;
				}
			}

			// look for 2nd backmost player
			if (m_iSide == RIGHT_TO_LEFT)
				dOpponentX = m_pOtherTeam[i]->GetAbsX();
			else
				dOpponentX = -m_pOtherTeam[i]->GetAbsX();
			if (dOpponentX < d2BackmostX)
			{
				if (dOpponentX < dBackmostX)
				{
					// backmost player
					d2BackmostX = dBackmostX;
					dBackmostX = dOpponentX;
					m_p2BackmostOpponent = m_pBackmostOpponent ;
					m_pBackmostOpponent = m_pOtherTeam[i];
				}
				else
				{
					d2BackmostX = dOpponentX;
					m_p2BackmostOpponent = m_pOtherTeam[i];
				}
			}
		}
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::KnowOffside()
Date    : 31.07.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the player is in offside
Description: called, when the player has seen complete field to other goal, so 
				 he should know about offside. This function then decides if he is 
				 in offside
-----------------------------------------------------------------------------*/
BOOL TPerception::KnowOffside()
{
	if (!m_bKnowNotInOffside)
		// player is not sure if in offside or not, now he can be: offside
		m_dInOffside = 1.0;

	m_iCheckOffsideCounter = 0;
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::CheckBallForOffside()
Date    : 14.04.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: decays offside truth value and checks if ball prevents agent from
				 beeing in offside
-----------------------------------------------------------------------------*/
void TPerception::CheckBallForOffside()
{
	m_dInOffside *= .98;
	m_bKnowNotInOffside = FALSE;

	// check if ball prevents offside
	if (m_iSide == LEFT_TO_RIGHT)	
	{
		if (m_pBall->GetAbsX() - m_pThisPlayer->GetAbsX() > 0)
		{
			m_bKnowNotInOffside = TRUE;
			m_dInOffside = 0.0;
			m_iCheckOffsideCounter = 0;
		}
	}	
	else
	{
		if (m_pBall->GetAbsX() - m_pThisPlayer->GetAbsX() < 0)
		{
			m_bKnowNotInOffside = TRUE;
			m_dInOffside = 0.0;
			m_iCheckOffsideCounter = 0;
		}
	}	
}

/*-----------------------------------------------------------------------------
Function: TPlayer* TPerception::GetAnyPlayerAtBall()
Date    : 27.05.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: the teammate or opponent, that is visible and is closest to this
Description: compares the two InFront players in distance and takes the distance
				 of this agent to the ball into account
-----------------------------------------------------------------------------*/
TPlayer* TPerception::GetAnyPlayerAtBall()
{
	if (m_pTeammateAtBall == NULL)
		return m_pOpponentAtBall;
	if (m_pOpponentAtBall == NULL)
		return m_pTeammateAtBall;
	if (GetObjectDistance(m_pTeammateAtBall) < GetObjectDistance(m_pOpponentAtBall))
		return m_pTeammateAtBall;
	else
		return m_pOpponentAtBall;
}


/*-----------------------------------------------------------------------------
Function: void TPerception::SetStamina(int stamina)
Date    : 27.08.98
-------------------------------------------------------------------------------
Parameter: stamina of the agent
Returnval: (void)
Description: sets the stamina and the stamina of thisPlayer for graphics
-----------------------------------------------------------------------------*/
void TPerception::SetStamina(int stamina)
{
	m_iStamina = stamina;
	m_pThisPlayer->SetStamina(stamina);

	// check if we get see messages
	// a server error sometimes prevents agents with 180 view from getting see
	if ((m_iSenseTime - m_iSeeTime > 20) && m_pAgent->IsConnected())
		m_pAgent->ChangeView(90, 1);
}

/*-----------------------------------------------------------------------------
Function: void TPerception::GetListVisiblePlayers(CObList & pList)
Date    : 13.05.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: list start pointer
Description: creates a list of all visible players
-----------------------------------------------------------------------------*/

void TPerception::GetListVisiblePlayers(CObList & pList)
{
	int		i;

	for (i = 0; i < m_iNoOfPlayers; i++)
	{
		Insert(pList,m_pVisiblePlayer[i]);
	}
}


/*-----------------------------------------------------------------------------
Function: CString TPerception::GetValidSeen()
Date    : 11.06.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: - 
Description: creates a string with number of objects:
  ball, own and other players position who are valid. divided by " ".
-----------------------------------------------------------------------------*/

CString TPerception::GetValidSeen()
{
	CString	sMsg;
	CString	sStr1;
	CString	sStr2;
	char  	cSep = ' ';
	int		i;
	int		j = 0;
	int		iNum;
	double	xPos;
	double	yPos;

	//position for the ball - object number 0
	if (m_pBall->IsValid())
	{
		iNum = 0;
		xPos = m_pBall->GetAbsX();
		yPos = m_pBall->GetAbsY();

		sStr2.Format("%d %.1f %.1f ", iNum, xPos, yPos);
		sStr1 += sStr2;
		j++;
	}

	// position from dispatcher
	iNum = m_pThisPlayer->GetNum();
	xPos = m_pThisPlayer->GetAbsX();
	yPos = m_pThisPlayer->GetAbsY();

	sStr2.Format("%d %.1f %.1f ", iNum, xPos, yPos);
	sStr1 += sStr2;

	j++;

	//position of all seen players - own and other
	for (i = 0; i < m_iNoOfPlayers; i++)
	{
		iNum = m_pVisiblePlayer[i]->GetNum();
		if( iNum != NOT_EXISTING )
		{
			xPos = m_pVisiblePlayer[i]->GetAbsX();
			yPos = m_pVisiblePlayer[i]->GetAbsY();

			if( !m_pVisiblePlayer[i]->IsOwnTeam() )
				iNum += 20;

			sStr2.Format("%d %.1f %.1f ", iNum, xPos, yPos);
			sStr1 += sStr2;
			j++;
		}
	}

	sStr2.Format("%d ",j);
	sMsg = sMsg + sStr2;
	sMsg = sMsg + sStr1;

	return sMsg;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::OnDashed(int power)
Date    : 07.07.99
-------------------------------------------------------------------------------
Parameter: power with which the player kicked
Returnval: (void)
Description: called when the player sent a dash command
-----------------------------------------------------------------------------*/
void TPerception::OnDashed(int power)
{
	UpdatePerceptionAfterDash(power);
	m_iExpectedNoOfDash = m_iNoOfDash + 1;
	m_iDashPower = power;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::OnKicked(int iPower, int iDirection)
Date    : 29.06.99
-------------------------------------------------------------------------------
Parameter: power (0 .. 100)
			  relative direction to kick
Returnval: (void)
Description: updates the ball
-----------------------------------------------------------------------------*/
void TPerception::OnKicked(int iPower, int iDirection)
{
	UpdatePerceptionAfterKick(iPower, iDirection);
	m_iExpectedNoOfKick = m_iNoOfKick + 1;
	m_iKickPower = iPower;
	m_iKickDirection= iDirection;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::OnTurned(int dir)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle that has been turned
Returnval: (void)
Description: called after the agent executed a turn command
-----------------------------------------------------------------------------*/
void TPerception::OnTurned(int dir)
{
	UpdatePerceptionAfterTurn(dir);
	m_iExpectedNoOfTurn = m_iNoOfTurn + 1;
	m_iTurnAngle = dir;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::OnTurnedNeck(int dir)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle that has been turned
Returnval: (void)
Description: called after the agent executed a turn_neck command
-----------------------------------------------------------------------------*/
void TPerception::OnTurnedNeck(int dir)
{
	UpdatePerceptionAfterTurnNeck(dir);
	m_iExpectedNoOfTurnNeck = m_iNoOfTurnNeck + 1;
	m_iTurnAngleNeck = dir;
	m_iHeadAngle += dir;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::UpdatePerceptionAfterDash()
Date    : 07.07.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: updates the agent's speed and position according to the executed dash command
-----------------------------------------------------------------------------*/
void TPerception::UpdatePerceptionAfterDash(int power)
{
	double	dSpeed;

	if (power < 0)
		power = -power;
	if (power > 100)
		power = 100;

	// update position after a dash is done once sense body is received
	// but since sense_body passes the decayed speed, we have to update the position
	// for the not decayed part.
	dSpeed = DASH_POWER_RATE * GetEffort() * power;
	dSpeed -= (dSpeed + GetSpeed()) * PLAYER_DECAY - GetSpeed();
	m_pThisPlayer->SetPushSpeed(dSpeed);
}

/*-----------------------------------------------------------------------------
Function: void TPerception::UpdatePerceptionAfterKick(int iPower, int iDirection)
Date    : 29.06.99
-------------------------------------------------------------------------------
Parameter: power and relative direction of the kick
Returnval: (void)
Description: updates the ball speed according to the executed kick command
-----------------------------------------------------------------------------*/
void TPerception::UpdatePerceptionAfterKick(int iPower, int iDirection)
{
	// we assume, that the kick has been successfull
	TVector	newSpeed;
	TVector	ballSpeed = m_pBall->GetAbsSpeed();
	double	dBallDist = GetObjectDistance(m_pBall) - PLAYER_SIZE - BALL_SIZE;
	double	dDeltaDir = (double) GetObjectDirection(m_pBall) / 180.0;
	double	dAcceleration;
	int		iDir = iDirection + m_pThisPlayer->GetAbsBodyDir();

	if (dBallDist > KICKABLE_MARGIN)
		// the ball is too far to kick
		return;

	if (dDeltaDir < 0)
		dDeltaDir = -dDeltaDir;

	// calculate the acceleration the ball is expected to get
	dAcceleration = iPower * KICK_POWER_RATE 
							* (1.0 - 0.25 * dDeltaDir - 0.25 * dBallDist / KICKABLE_MARGIN);
	
	newSpeed.SetVectorByDirection(dAcceleration, iDir);
	ballSpeed += newSpeed;
	// set absolute ball speed
	m_pBall->SetAbsSpeed(ballSpeed);

	// ball position is updated next UpdatePerception
}

/*-----------------------------------------------------------------------------
Function: void TPerception::UpdatePerceptionAfterTurn(int angle)
Date    : 23.06.99
-------------------------------------------------------------------------------
Parameter: angle the agent turned or turned_neck
Returnval: (void)
Description: all objects' directions except lines and flags are updated
-----------------------------------------------------------------------------*/
void TPerception::UpdatePerceptionAfterTurn(int angle)
{
	TDirection	direction;
	
	// update the agent's absolute body direction
	direction = m_pThisPlayer->GetAbsBodyDir();
	direction += angle;
	m_pThisPlayer->SetAbsBodyDirection(direction);

	UpdatePerceptionAfterTurnNeck(angle);
}

/*-----------------------------------------------------------------------------
Function: void TPerception::UpdatePerceptionAfterTurnNeck(int angle)
Date    : 23.06.99
-------------------------------------------------------------------------------
Parameter: angle the agent turned or turned_neck
Returnval: (void)
Description: all objects' directions except lines and flags are updated
-----------------------------------------------------------------------------*/
void TPerception::UpdatePerceptionAfterTurnNeck(int angle)
{
	int			i;
	TDirection	direction;
	
	// update the agent's absolute view direction
	direction = m_pThisPlayer->GetAbsDir();
	direction += angle;
	m_pThisPlayer->SetAbsDirection(direction);

	// update visible players
	for (i = 0; i < m_iNoOfPlayers; i++)
		m_pVisiblePlayer[i]->UpdateDirection(angle);

	// update remembered players (global map)
	for (i = 0; i < PLAYERS_PER_TEAM; i++)
	{
		m_pOwnTeam[i]->UpdateDirection(angle);
		m_pOtherTeam[i]->UpdateDirection(angle);
	}

	// and other objects
	m_pOwnGoal->UpdateDirection(angle);
	m_pOtherGoal->UpdateDirection(angle);
	m_pBall->UpdateDirection(angle);
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::IsOutside(TVector& pos)
Date    : 24.06.99
-------------------------------------------------------------------------------
Parameter: position that is checked
Returnval: true, if the position is not within the field
Description: checks the position to be outside or inside field
-----------------------------------------------------------------------------*/
BOOL TPerception::IsOutside(TVector& pos)
{
	double sizeX = PITCH_LENGTH / 2.0;
	double sizeY = PITCH_WIDTH / 2.0;

	if ((pos.GetX() < -sizeX) || (pos.GetX() > sizeX)
		 || (pos.GetY() < -sizeY) || (pos.GetY() > sizeY))
		return TRUE;
	else
		return FALSE;
}
