// TVector.cpp: implementation of the TVector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Basis.h"
#include "TDirection.h"
#include "TVector.h"

#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TVector::~TVector()
{
}

void TVector::SetVectorByDirection(double len, int dir)
{
	double direction = (double)dir * DEG2RAD;
	m_dX = len * cos(direction);
	m_dY = len * sin(direction);
}

#ifdef _DEBUG
void TVector::Dump(CDumpContext& dc) const
{
	CString	text;
	text.Format("x: %4.2f y: %4.2f len: %4.2f ", m_dX, m_dY, GetLength());
	dc << text;
}
#endif

/*-----------------------------------------------------------------------------
Function: void TVector::Rotate(int degrees)
Date    : 02.04.98
-------------------------------------------------------------------------------
Parameter: angle to rotate in -180 - 180 degrees
Returnval: (void)
Description: rotates the vector by the given amount
-----------------------------------------------------------------------------*/
void TVector::Rotate(int degrees)
{
	double	rad = (double) degrees * DEG2RAD;
	double	sinW = sin(rad);
	double	cosW = cos(rad);
	double	x;

	x = m_dX * cosW + m_dY * sinW;
	m_dY = m_dX * -sinW + m_dY * cosW;
	m_dX = x;
}

/*-----------------------------------------------------------------------------
Function: TDirection TVector::GetDirection() const
Date    : 04.04.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: direcetion, the vector points to in degrees. 0 degrees is to positive
				x axis, 90 degrees is the positive y axis
Description: 
-----------------------------------------------------------------------------*/
TDirection TVector::GetDirection() const
{
	TDirection	angle;

	if ((m_dX > D_ZERO_N) && (m_dX < D_ZERO_P))
		if (m_dY > 0.0)
			angle = 90;
		else
			angle = -90;
	else
		if (m_dX > 0.0)
			angle = atan(m_dY / m_dX) * RAD2DEG;
		else
			angle = atan(m_dY / m_dX) * RAD2DEG - 180.0;

	return angle;
}

/*-----------------------------------------------------------------------------
Function: double TVector::GetLength() const
Date    : 04.04.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: length of the vector (standar scalar product)
Description: returns the length of a vector
-----------------------------------------------------------------------------*/
double TVector::GetLength() const
{
	return sqrt(m_dX * m_dX + m_dY * m_dY);
}

TVector TVector::operator +(TVector& vec2)
{
	return TVector(m_dX + vec2.GetX(), m_dY + vec2.GetY());
}

TVector TVector::operator -(TVector& vec2)
{
	return TVector(m_dX - vec2.GetX(), m_dY - vec2.GetY());
}

TVector& TVector::operator +=(TVector& vec2)
{
	m_dX += vec2.GetX();
	m_dY += vec2.GetY();
	return *this;
}

TVector& TVector::operator -=(TVector& vec2)
{
	m_dX -= vec2.GetX();
	m_dY -= vec2.GetY();
	return *this;
}

TVector& TVector::operator *=(double val)
{
	m_dX *= val;
	m_dY *= val;
	return *this;
}

TVector operator*(double val1, TVector& val2) { return TVector(val1 * val2.GetX()
																					,val1 * val2.GetY()); }
TVector operator*(TVector& val1, double val2) { return TVector(val2 * val1.GetX()
																					,val2 * val1.GetY()); }
