// TViewingAgent.h: interface for the TViewingAgent class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TVIEWINGAGENT_H__0F8D1D81_762C_11D1_B242_0000C09B5CBE__INCLUDED_)
#define AFX_TVIEWINGAGENT_H__0F8D1D81_762C_11D1_B242_0000C09B5CBE__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "BaseAgent.h"


#ifdef _DEBUG
	// define to see some perception variables in output window
	#define xSHOW_VIEWING_VAR
#endif

// positions for the players of left team at kick on(right team is -x, y and right left exchanged)
/* Version 2 KI 98
#define POS_KICKOFF_LEFT_X { -50, -10, -10, -10,  -8, -9, -8,  -1,  -1, -1,  -1 }
#define POS_KICKOFF_LEFT_Y {   0, -15,   3,  15, -23,  -6, 23,  -20, -1,  7,  20 }
#define POS_KICKOFF_RIGHT_X { -50, -10, -10, -10,  -8, -9,  -8,  -1,  -1, -1, -1 }
#define POS_KICKOFF_RIGHT_Y {   0, -15,   3,  15, -23, -6,  23, -20,  -9,  9, 20 }
#define POS_HOME_X { -50, -10, -10, -10,  -8, -9, -8,  20,  30, 30,  20 }
#define POS_HOME_Y {   0, -15,   3,  15, -23, -6, 23,  -20, -7,  7,  20 }
*/
// Version 11 players
#define POS_KICKOFF_LEFT_X { -48, -14, -14, -14, -14, -10, -3, -10,  -1, -1,  -1 }
#define POS_KICKOFF_LEFT_Y {   0,  -15, -5,   5,  15, -20,  0,  20, -15,  0,  15 }
#define POS_KICKOFF_RIGHT_X { -48, -14, -14, -14, -14, -10, -9, -10,  -1,  -1,  -1 }
#define POS_KICKOFF_RIGHT_Y {   0, -15,  -5,   5,  15, -20,  0,  20, -15,  -9,  15 }
#define POS_HOME_X { -48, -14, -14, -14, -14, -10, -3, -10,  20, 30, 20 }
#define POS_HOME_Y {   0, -15,  -5,   5,  15, -20,  0,  20, -15,  0, 15 }


/* // test version
#define POS_KICKOFF_LEFT_X { -48, -48, -48, -14, -14, -10, -3, -10,  -1, -1,  -1 }
#define POS_KICKOFF_LEFT_Y {   0,  10, -10,   5,  15, -20,  0,  20, -15,  0,  15 }
#define POS_KICKOFF_RIGHT_X { -48, -48, -48, -14, -14, -10, -9, -10,  -1,  -1,  -1 }
#define POS_KICKOFF_RIGHT_Y {   0, 10,  -10,   5,  15, -20,  0,  20, -15,  -9,  15 }
#define POS_HOME_X { -48, -48, -48, -14, -14, -10, -3, -10,  20, 30, 20 }
#define POS_HOME_Y {   0, 10,  -10,   5,  15, -20,  0,  20, -15,  0, 15 }
*/

/*
// Version for 2 players
#define POS_KICKOFF_LEFT_X { -10, -10, -14, -14, -14, -10, -3, -10,  -1, -1,  -1 }
#define POS_KICKOFF_LEFT_Y {   0,  -10, -5,   5,  15, -20,  0,  20, -15,  0,  15 }
#define POS_KICKOFF_RIGHT_X { -10, -10, -14, -14, -14, -10, -9, -10,  -1,  -1,  -1 }
#define POS_KICKOFF_RIGHT_Y {   0, -10,  -5,   5,  15, -20,  0,  20, -15,  -9,  15 }
#define POS_HOME_X { -10, -10, -14, -14, -14, -10, -3, -10,  20, 30, 20 }
#define POS_HOME_Y {   0, -10,  -5,   5,  15, -20,  0,  20, -15,  0, 15 }
*/
class TPerception;
class TScreenParams;
class CBasisView;
class	TPerceptionView;
class TNetworkView;
class TRuleView;
class TVisibleObject;
class TDirection;
class TBehaviorNetwork;

/*-----------------------------------------------------------------------------
Class   : TViewingAgent
Date    : 15.12.97
-------------------------------------------------------------------------------
Base Class: TBaseAgent

Description: extends the base agents capabilities to perception
-----------------------------------------------------------------------------*/
class TViewingAgent : public CBaseAgent  
{
	public:
		TViewingAgent(CBasisView* pBView, TPerceptionView* pPView
							, TNetworkView* pNView, TRuleView* pRView);
		virtual ~TViewingAgent();

		// normal methods
		void	SwitchViewAngle(void); 
		void	SwitchViewQuality(void);
		void	SetWaitForNextSee(BOOL val = TRUE) { m_bWaitForNextSee = val; };
		BOOL	GetWaitForNextSee() const { return m_bWaitForNextSee; };
		double	GetLinearFuzzyValue(double value, double left, double right, int up = TRUE);

		// network functions
		TNetworkParams*	GetNetworkParams();
		int					GetNoOfRules() const;
		virtual void		SetKnowledgeType(int type);
		int					GetKnowledgeType(void) const { return m_iKnowledgeType; };
		void					ResetNetwork();

		// perception functions
		TPerception*	GetPerception() { return m_pPerception; };
		int				GetSide();
		void				SetSide(int side);
		virtual void	SetNumber(int number, int side);
		CPoint			GetPosKickOffLeft() const { return m_iPosKickOffLeft; };
		CPoint			GetPosKickOffRight() const { return m_iPosKickOffRight; };
		CPoint&			GetPosHome() { return m_iPosHome; };
		void				SetPosHome(CPoint &point) { m_iPosHome = point; };
		int				GetStamina() const;

		void				SetOrdinaryGame(BOOL bVal) { m_bOrdinaryGame = bVal; };
		BOOL				GetOrdinaryGame(void) const { return m_bOrdinaryGame; };

		// overridables
		virtual void OnConnected();
		virtual void OnDisconnect();
		virtual void Turn(int iMoment, BOOL bQueue = FALSE);
		void TurnBody(int iMoment, BOOL bQueue = FALSE);

		virtual void OnSentMessage(LPCTSTR message);
		virtual void OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork = TRUE);
		virtual void OnReceivedSee() {};
		virtual void OnReceivedLastSenseBody() {};
			
		virtual void OnMoved(int x, int y);
		virtual void OnTurned(int dir);
		virtual void OnDashed(int pow);
		virtual void OnKicked(int pow, int dir);
		virtual void OnCaught(int dir);
		virtual void OnChangedView(int mode, int quality);
		virtual void OnTurnedNeck(int dir);


		// perception functions ordered alphabetically -----------
		void		CreatePerceptions();

		double	GetBallAhead(void);
		double	GetBallBehind(void);
		double	GetBallFarer(void);
		double	GetBallIsCatchable(void);
		double	GetBallIsFar(void);
		double	GetBallIsNearOtherGoal(void);
		double	GetBallIsNearOwnGoal(void);
		double	GetBallIsVeryNearOwnGoal(void);
		double	GetBallKickable(void);
		double	GetBallNearer(void);
		double	GetBallOutside(void);
		double	GetCatchSuccess(void);
		double	GetFalse(void);
		double	GetFreeTeammateIsOffside(void);
		double	GetHaveBall(void);
		double	GetHaveEnoughStamina(void);
		double	GetHaveSeenBall(void);
		double	GetHaveSeenTeammate(void);
		double	GetHighStamina(void);
		double	GetInOffside(void);
		double	GetInOtherHalf(void);
		double	GetJustCaughtBall(void);
		double	GetKnowNotInOffside(void);
		double	GetLowStamina(void);
		double	GetMaySaySomething(void);
		double	GetNearBall(void);
		double	GetNearGoal(void);
		double	GetNearOpponent(void);
		double	GetNearOwnGoal(void);
		double	GetNoOtherGoal(void);
		double	GetOffsideTrap(void);
		double	GetOpponentHasBall(void);
		double	GetOpponentInFront(void);
		double	GetOpponentIsNearerToBall(void);
		double	GetOpponentIsNearerToOtherGoal(void);
		double	GetOpponentIsNearerToOwnGoal(void);
		double	GetOtherGoalNearer(void);
		double	GetOwnGoalNearer(void);
		double	GetPlaymodeOn(void);
		double	GetSeeBall(void);
		double	GetSoccergoal(void);
		double	GetStandFree(void);
		double	GetTeammateBehind(void);
		double	GetTeammateHasBall(void);
		double	GetTeammateInFront(void);
		double	GetTeammateIsBetweenBall(void);
		double	GetTeammateIsFree(void);
		double	GetTeammateIsNearerBall(void);
		double	GetTeammateIsNearerOtherGoal(void);
		double	GetTrue(void);
				
		// -------------------------------------------------------
	
		// resource functions		
		void		CreateResources();
		double	GetLegsAvailable();
		double	GetMouthsAvailable();
		double	GetNecksAvailable();
		double	GetStaminaAvailable();

	protected:

		TBehaviorNetwork*	m_pBehaviorNetwork;
		TNetworkView*		m_pNetworkView;
		TRuleView*			m_pRuleView;
		int					m_iKnowledgeType;

		TPerception*	m_pPerception; // perception class of the agent
		CBasisView*		m_pView;			// view to display network traffic

		CPoint			m_iPosKickOffLeft;
		CPoint			m_iPosKickOffRight;
		CPoint			m_iPosHome;
		int				m_iIndexPosHome;	// position index for goalie
		BOOL				m_bOrdinaryGame;
		BOOL				m_bWaitForNextSee;
};

#endif // !defined(AFX_TVIEWINGAGENT_H__0F8D1D81_762C_11D1_B242_0000C09B5CBE__INCLUDED_)
