// TVisibleObject.cpp: implementation of the TVisibleObject class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "Basis.h"
#include "TScreenParams.h"
#include "TVisibleObject.h"

#include	<math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


/*-----------------------------------------------------------------------------
Function: TVisibleObject::TVisibleObject()
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: none
Description: constructor
-----------------------------------------------------------------------------*/
TVisibleObject::TVisibleObject()
{
	m_iTime = -1000;
	m_iOldTime = m_iTime;
	m_dDistance = 0.0;
	m_iDirection = 0;
	m_dDeltaDist = 0.0;
	m_iDeltaDir = 0;
	m_bValid = FALSE;
	m_bInMind = FALSE;
	m_bOnceNotSeen = FALSE;
}

TVisibleObject::~TVisibleObject()
{
}

/*-----------------------------------------------------------------------------
Function: void GetCoordinates(int &x, int &y, TScreenParams *tScreen)
Date    : 14.11.97
-------------------------------------------------------------------------------
Parameter: reference to the coordinates to return
Returnval: (void)
Description: calculates screen coordinates from distance and direction
-----------------------------------------------------------------------------*/
void TVisibleObject::SetCoordinates(TScreenParams *tScreen)
{
	double direction = ((double) (m_iDirection - 90)) * DEG2RAD;
	// rad is x * pi / 180
	if (tScreen->IsTopView())
	{
		// the perception is drawn from a top view
		m_iX = tScreen->GetMitteX() + (int) (tScreen->GetScale() * m_dDistance * cos(direction)); 	
		m_iY = tScreen->GetMitteY() + (int) (tScreen->GetScale() * m_dDistance * sin(direction)); 	
	}
	else if (tScreen->IsRelativeView())
	{
		double x;
		double y;
		// it is drawn as seen by the player
		x = m_dDistance * cos(direction); 	
		y = - m_dDistance * sin(direction); 	

		// m_iX = x * distance of viewplane / y
		if (y > 0.01)
			m_iX = tScreen->GetMitteX() + (int) (x * VIEWPLANE_DIST * tScreen->GetScale()/ y);
		else
			m_iX = 1000;

		// m_iY = distance to view plane * height of eye / distance of object 
		if (m_dDistance > 0.01)
			m_iY = tScreen->GetMitteY() + (int) (VIEWPLANE_DIST * tScreen->GetScale()
															* EYE_HEIGHT / m_dDistance);
		else
			m_iY  = 1000;
	}
	else
	{
		// absolute perception and world model
		if ((tScreen->GetTime() > 3000) && (tScreen->GetTime() <= 6000))
		{
			// second half
			m_iX = (int) (((double) -m_tPosition.GetX() + PITCH_MARGIN + PITCH_LENGTH / 2.0)
								* tScreen->GetAbsScale());
			m_iY = (int) (((double) -m_tPosition.GetY() + PITCH_MARGIN + PITCH_WIDTH / 2.0)
								* tScreen->GetAbsScale());
		}
		else
		{
			// first half or extended time
			m_iX = (int) (((double) m_tPosition.GetX() + PITCH_MARGIN + PITCH_LENGTH / 2.0)
								* tScreen->GetAbsScale());
			m_iY = (int) (((double) m_tPosition.GetY() + PITCH_MARGIN + PITCH_WIDTH / 2.0)
								* tScreen->GetAbsScale());
		}
	}
}

/*-----------------------------------------------------------------------------
Function: void TVisibleObject::SetVisibleObject(TVisibleObject* obj)
Date    : 12.03.98
-------------------------------------------------------------------------------
Parameter: pointer to the object to be copied
Returnval: (void)
Description: copies the passed object to this
-----------------------------------------------------------------------------*/
void TVisibleObject::SetVisibleObject(TVisibleObject* obj)
{
	m_iTime = obj->GetTime();
	m_iOldTime = obj->GetOldTime();
	m_dDistance = obj->GetDistance();
	m_iDirection = obj->GetDirection();
	m_dDeltaDist = obj->GetDeltaDist();
	m_iDeltaDir = obj->GetDeltaDir();
	m_bValid = obj->IsValid();
	m_tPosition = obj->GetPosition();
	m_bInMind = obj->IsInMind();
	m_bOnceNotSeen = obj->IsOnceNotSeen();
}

/*-----------------------------------------------------------------------------
Function: void TVisibleObject::SetVisibleObject(int time, double dist, int dir)
Date    : 12.03.98
-------------------------------------------------------------------------------
Parameter: time, distance and relative direction of the object
Returnval: (void)
Description: copies the passed object to this
-----------------------------------------------------------------------------*/
void TVisibleObject::SetVisibleObject(int time, double dist, int dir)
{
	m_iOldTime = m_iTime;
	m_iTime = time;
	SetDistance(dist);
	SetDirection(dir);
	m_bValid = TRUE;
	m_bInMind = TRUE;
	m_bOnceNotSeen = FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TVisibleObject::SetDistance(double dist)
Date    : 21.04.98
-------------------------------------------------------------------------------
Parameter: new relative distance of the object
Returnval: (void)
Description: calculates the change in distance, sets the new distance and then
				 sets the object as valid (seen this cycle)
-----------------------------------------------------------------------------*/
void TVisibleObject::SetDistance(double dist)
{ 
	// calculate the change in distance
	m_dDeltaDist = dist - m_dDistance;
	m_dDistance = dist; 
	m_bValid = TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TVisibleObject::SetDirection(TDirection dir)
Date    : 21.04.98
-------------------------------------------------------------------------------
Parameter: new relative direction of the object
Returnval: (void)
Description: calculates the change in direction and sets the new direciton
-----------------------------------------------------------------------------*/
void TVisibleObject::SetDirection(TDirection dir)
{ 
	// calculate the change in direction
	m_iDeltaDir = dir - m_iDirection;
	m_iDirection = dir; 
}

/*-----------------------------------------------------------------------------
Function: double TVisibleObject::GetDistanceTo(double x, double y) const
Date    : 02.04.98
-------------------------------------------------------------------------------
Parameter: absolute coordinates of a point to which the distance is calculated
Returnval: distance from this object to the passed point
Description: calculates the object's distances using their absolute coordinates
-----------------------------------------------------------------------------*/
double TVisibleObject::GetDistanceTo(TVector& pos)
{
	TVector	delta(m_tPosition - pos);
	double	dx = delta.GetX();
	double	dy = delta.GetY();
	return sqrt(dx*dx + dy*dy);
}

/*-----------------------------------------------------------------------------
Function: void TVisibleObject::SetPosition(TVector& pos)
Date    : 06.07.99
-------------------------------------------------------------------------------
Parameter: the position for the object
Returnval: (void)
Description: sets the objects position and stores the previous position
-----------------------------------------------------------------------------*/
void TVisibleObject::SetPosition(TVector& pos)
{
	m_tOldPos = m_tPosition;
	m_tPosition = pos;
}

/*-----------------------------------------------------------------------------
Function: TMoveableObject::TMoveableObject()
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: none
Description: 
-----------------------------------------------------------------------------*/
TMoveableObject::TMoveableObject()
{
	m_dDistChng = 0.0;
	m_dDirChng = 0.0;
	Initialize();
}

void TMoveableObject::Initialize()
{
	m_iTime = 0;
	m_bOnceSeen = FALSE;
	m_bInMind = FALSE;
	m_tPosition.SetVector(-1000.0, -1000.0);
}
 

/*-----------------------------------------------------------------------------
Function: void TMoveableObject::SetMoveableObject(int time, double dist, int dir, double distchng, int dirchng)
Date    : 01.04.98
-------------------------------------------------------------------------------
Parameter: parameters of a moveable object
Returnval: (void)
Description: sets the parameters
-----------------------------------------------------------------------------*/
void TMoveableObject::SetMoveableObject(int time, double dist, int dir, double distchng, double dirchng)
{
	// set the new parameters
	SetVisibleObject(time, dist, dir);
	m_dDistChng = distchng;
	m_dDirChng = dirchng;
	m_bOnceSeen = TRUE;
}

void TMoveableObject::SetMoveableObject(TMoveableObject *obj)
{
	SetVisibleObject(obj); 
	m_bOnceSeen = TRUE;
	m_tAbsSpeed = obj->GetAbsSpeed();
//	m_tRelSpeed = obj->GetRelSpeed();
	m_dDistChng = obj->GetDistChng();
	m_dDirChng = obj ->GetDirChng();
	m_tLastSeePosition = obj->GetLastSeePosition();
}

/*-----------------------------------------------------------------------------
Function: void TMoveableObject::UpdatePosition(int iDataAge, iSeeTime)
Date    : 23.03.98
-------------------------------------------------------------------------------
Parameter: indicates, how much action cycles (100ms) have been since last perception
			  time of latest see message needed to calculate the age of perception of 
			  this object
Returnval: (void)
Description: updates the position (direction and distance) of an object
				 according to distchange, direction change in a linear way. This
				 function is called every 100 ms before an action has to be selected
-----------------------------------------------------------------------------*/
void TMoveableObject::UpdatePosition(int iDataAge, int iSeeTime)
{
	if (IsValid() && (iDataAge == 0))
		// the object is visible and has just be seen
		return;
	
	// the object is visible, but perception is slower than action, so update
	// also update invisible object's absolute coordinates
	SetPosition(m_tPosition + m_tAbsSpeed);
	m_tAbsSpeed *= GetDecay();

	if (!IsValid() && (iSeeTime - m_iTime >= MAX_IMAGINE_TIME))
		m_bInMind = FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TMoveableObject::SetAbsCoordinates(TMoveableObject* pAgent, int dir)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: pointer to the agent
			  absolute direction the agent is facing
Returnval: (void)
Description: calculates the new absolute position of this object according to
				 the absolute position of the agent (x, y) and the relative distance
				 and direction in which this object is perceived
-----------------------------------------------------------------------------*/
void TMoveableObject::SetAbsCoordinates(TMoveableObject* pAgent, int dir)
{
	double	direction = ((double)(dir + m_iDirection)) * DEG2RAD;
	TVector	tNewPos = pAgent->GetPosition() 
								+ TVector(cos(direction), sin(direction)) * m_dDistance;
	double	distance1 = (GetPosition() - tNewPos).GetLength();
	double	distance2 = (GetLastSeePosition() - tNewPos).GetLength();
	BOOL		bWhich;

	// check if the new position is closer to the expected position or to the
	// latest see position
	if (distance1 < distance2)
	{
		if (distance1 < 5.0)
			// only filter if positions are rather close
			tNewPos = 0.5 * (m_tPosition + tNewPos);
		bWhich = TRUE;
	}
	else
	{
		// ball is closer to last view position
		if (distance2 < 5.0)
			// only filter if positions are rather close
			tNewPos = 0.3 * GetLastSeePosition() + 0.7 * tNewPos;
		bWhich = FALSE;
	}

	// and the new position
	SetPosition(tNewPos);	
	// set the relative and absolute speed
	SetSpeed(pAgent, dir, bWhich);
	// remember the last position where the object has been seen
	SetLastSeePosition();
}

/*-----------------------------------------------------------------------------
Function: TVector TMoveableObject::GetRelSpeed(TVector& pos, int dir)
Date    : 29.06.99
-------------------------------------------------------------------------------
Parameter:	absolute position of the agent
				absolute body direction of the agent
Returnval: (void)
Description: calculate relative speed values
-----------------------------------------------------------------------------*/
TVector TMoveableObject::GetRelSpeed(TVector& pos, int dir)
{
	// calculate according to the absolute direction the agent is facing
	TVector posObject = m_tPosition;
	TVector posSpeed = posObject + m_tAbsSpeed;

	// translate to agents coordinate system
	posObject -= pos;
	posObject.Rotate(dir);
	posSpeed -= pos;
	posSpeed.Rotate(dir);
	return posSpeed - posObject;
}

/*-----------------------------------------------------------------------------
Function: void TMoveableObject::SetSpeed(TMoveableObject* pAgent, int dir, BOOL bWhich)
Date    : 26.03.98
-------------------------------------------------------------------------------
Parameter: pointer to this agent
			  absolute direction the agent is facing
			  flag indicating, if the oldPosition (TRUE) or lastSeePosition is used
Returnval: (void)
Description: calculates the new absolute speed of this object according to
				 the absolute position of the agent (x, y) and the relative distance
				 and direction in which this object is perceived
-----------------------------------------------------------------------------*/
void TMoveableObject::SetSpeed(TMoveableObject* pAgent, int dir, BOOL bWhich)
{
	double	dMaxSpeed;
	double	dAbsSpeed;
	double	dRatio;

	// if distchng or dirchng is not existing use absolute coordinate change
	if ((m_dDistChng < NOT_EXISTING + 1.0) || (m_dDirChng < NOT_EXISTING + 1.0))
	{
		// take absolute coordinates change
		TVector	tAbsSpeed;

		if (bWhich)
			// take stored previous position
			tAbsSpeed = GetPosition() - GetOldPos();
		else
		{
			int	deltaTime = m_iTime - m_iOldTime;
			// take last see position
			tAbsSpeed = GetPosition() - m_tLastSeePosition;
		
			if (deltaTime > 1)
				tAbsSpeed *= 1.0 / (double) deltaTime;
		}

		// filter with previous speed
		m_tAbsSpeed = 0.3 * m_tAbsSpeed + 0.7 * tAbsSpeed;
/*		
		#ifdef _DEBUG
			afxDump << "ballspeed:" << tAbsSpeed << " after " << m_tAbsSpeed << "\n";
		#endif
*/
//		SetRelativeFromAbsoluteSpeed(pAgent->GetPosition(), dir);
	}
	else
	{
		// we have distance and dir change information
		TVector	tRelSpeed;
		double	dDirRad = (double) m_iDirection * PI / 180;
		double	dDirAlt = ((double) m_iDirection - m_dDirChng) * PI / 180;
		double	dDistAlt = m_dDistance - m_dDistChng;

		tRelSpeed = m_dDistance * TVector(cos(dDirRad), sin(dDirRad))
						- dDistAlt * TVector(cos(dDirAlt), sin(dDirAlt));
		// Filter the relative speed by the previous value, if changes are small
//		if ((tRelSpeed - m_tRelSpeed).GetLength() < 0.4)
//			tRelSpeed = (m_tRelSpeed + tRelSpeed) * 0.5;
//		m_tRelSpeed = tRelSpeed;

		// calculating absolute speed
		m_tAbsSpeed = tRelSpeed;
		m_tAbsSpeed.Rotate(-dir);
		m_tAbsSpeed += pAgent->GetAbsSpeed();
		// ??? Filter abs speed
	}

	// treat implausible values
	dMaxSpeed = GetSpeedMax();
	dAbsSpeed = m_tAbsSpeed.GetLength();
	if (dAbsSpeed > dMaxSpeed)
	{
		dRatio = dMaxSpeed / dAbsSpeed;
		m_tAbsSpeed *= dRatio;
//		m_tRelSpeed *= dRatio;
	}
}

/*-----------------------------------------------------------------------------
Function: void TMoveableObject::GetFuturePosition(double timeAhead, TVector& pos)
Date    : 02.04.98
-------------------------------------------------------------------------------
Parameter: no of timesteps to calculate ahead (in 100ms units)
			  pointer of the vector that gets the result
Returnval: (void)
Description: calculates the absolute position of this object due to its current 
				 position and the speed of the object
				 takes speed decay into account
-----------------------------------------------------------------------------*/
void TMoveableObject::GetFuturePosition(double timeAhead, TVector& pos)
{
	int		i;
	TVector	speed = m_tAbsSpeed;
	double	decay = GetDecay();

	pos = GetPosition();
	for (i = 0; i < timeAhead; i++)
	{
		pos += speed;
		speed *= decay;
	}
}

/*-----------------------------------------------------------------------------
Function: _
Date    : 21.04.98
-------------------------------------------------------------------------------
Parameter: _
Returnval: (void)
Description: _
-----------------------------------------------------------------------------*/
TPlayer::TPlayer()
{
	m_bThisAgent = FALSE;
	m_bMatched = FALSE;
	m_bOnlyInMind = FALSE;
	m_iHearTimePlayer = 0;
	m_iStamina = MAX_STAMINA;
}

/*-----------------------------------------------------------------------------
Function: void TPlayer::SetPlayer(int iTime, double dist, int dir, double distchng
							 , int dirchng, BOOL myteam, int team, int num)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: _
Returnval: (void)
Description: sets the players member variables
-----------------------------------------------------------------------------*/
void TPlayer::SetPlayer(int iTime, double dist, int dir, double distchng
							 , double dirchng, int faceDir, int bodyDir, BOOL myTeam, int team, int num)
{
	// sets the parameters of a player
	SetMoveableObject(iTime, dist, dir, distchng, dirchng);
	m_bOwnTeam = myTeam;
	m_iTeam = team;
	m_iNum = num;
}

/*-----------------------------------------------------------------------------
Function: void TPlayer::SetPlayer(TPlayer* player)
Date    : 12.03.98
-------------------------------------------------------------------------------
Parameter: pointer
Returnval: (void)
Description: sets the players member variables to the one passed
-----------------------------------------------------------------------------*/
void TPlayer::SetPlayer(TPlayer* player)
{
	// sets inherited params
	SetMoveableObject(player);
	
	// sets the parameters of a player
	m_bThisAgent = player->IsThisAgent();
	m_bOwnTeam = player->IsOwnTeam();
	m_iTeam = player->GetTeam();
	m_iNum = player->GetNum();
	
	// this function is called, when a seen player is matched to a player 
	// in the global map, so set the player to be matched with a player
	// in the global map
	SetMatched();
	SetOnlyInMind(FALSE);
}

/*-----------------------------------------------------------------------------
Function: BOOL TPlayer::CheckPreventOffside(TPlayer* other)
Date    : 31.07.98
-------------------------------------------------------------------------------
Parameter: pointer to a player of the opponent team
Returnval: TRUE, if the player passed is nearer to his own goal, or if this player
			  is in its own half which prevents this player to be in offside
Description: checks, if the other player prevents me from beeing offside
-----------------------------------------------------------------------------*/
BOOL TPlayer::CheckPreventOffside(TPlayer* other)
{
	if (IsInOwnHalfOfField())
		return TRUE;

	if (m_iTeam == LEFT_TO_RIGHT)
	{
		if (GetAbsX() <= other->GetAbsX())
			return TRUE;
	}
	else
	{
		if (GetAbsX() >= other->GetAbsX())
			return TRUE;
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TPlayer::DrawObject(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
Returnval: (void)
Description: draws the player as a circle (with the number in if existing)
-----------------------------------------------------------------------------*/
void TPlayer::DrawObject(CDC* pDC, TScreenParams *tScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	int		size;
	int		age;

	// draw only players having been percepted this cycle
	if (tScreen->IsAbsoluteView())
	{
		// do not draw players, that have never been seen before
		if (!GetOnceSeen())
			return;
	}
	else if (!IsValid())
		return;

	// what team is the player
	if (m_iTeam == LEFT_TO_RIGHT)
		cCol = COLOR_MYTEAM;
	else if (m_iTeam == RIGHT_TO_LEFT)
		cCol = COLOR_OTHERTEAM;
	else
		cCol = COLOR_ANY_OBJECT;

	if (tScreen->IsTopView())
		size = PLAYERSIZE;
	else if (tScreen->IsRelativeView())
	{
		if (m_dDistance > EYE_HEIGHT)
			size = (int) (PLAYERSIZE  * SCALE_DIST / m_dDistance);
		else
			size = (int) (PLAYERSIZE  * SCALE_DIST / EYE_HEIGHT);
	}
	else
	{
		// if the player is no longer in the mind of the object do not draw it
		if (!IsInMind())
			return;

		// absolute view
		age = tScreen->GetTime() - m_iTime;
		if (age < 0)
			age = 0;
		age *= 200 / MAX_IMAGINE_TIME;
		if (m_iTeam == LEFT_TO_RIGHT)
			cCol = RGB(200 - age, 200 - age, 20);
		else if (m_iTeam == RIGHT_TO_LEFT)
			cCol = RGB(200 - age, 20, 20);
		else
			cCol = COLOR_ANY_OBJECT;
		size = (int) (PLAYER_SIZE_DRAW * tScreen->GetAbsScale());
	}

	// set color
	brush.CreateSolidBrush(cCol);
	oldBrush = pDC->SelectObject(&brush);
	SetCoordinates(tScreen);

	// draw a Ellipse at the players position
	pDC->Ellipse(m_iX - size, m_iY - size, m_iX + size, m_iY + size);
	
	// draw degree of freeness
	if (tScreen->GetDrawDegreeOfFreeness())
	{
		int		nPointSize;
		CFont		font;
		CFont*	oldFont;
		CRect		field;
		COLORREF	oldTextColor;
		COLORREF	oldBkColor;
		int		sizeX;
		CString	text;

		nPointSize = (int) (20.0 * tScreen->GetAbsScale());
		if (!font.CreatePointFont(nPointSize, _T("Times New Roman"), pDC))
			font.CreateStockObject(SYSTEM_FONT);
		oldFont = pDC->SelectObject(&font);

		// calculate the rectangle of the text
		sizeX = (int) tScreen->GetScale() * nPointSize / 5;
		oldTextColor = pDC->SetTextColor(0x00000000);
		oldBkColor = pDC->SetBkColor(0xffffffff);

		// draw degree of freeness
		field.SetRect(m_iX, m_iY, m_iX + sizeX, m_iY + nPointSize / 5);
		GetObjectText(text);
		pDC->DrawText(text, field, DT_SINGLELINE);

		pDC->SetTextColor(oldTextColor);
		pDC->SetBkColor(oldBkColor);
		pDC->SelectObject(oldFont);
		font.DeleteObject();
	}

	pDC->SelectObject(oldBrush);
	brush.DeleteObject();
}

/*-----------------------------------------------------------------------------
Function: BOOL TPlayer::GetInLeftSideOfField()
Date    : 27.05.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the player is on the left side of the field (seen in playdirection)
Description: _
-----------------------------------------------------------------------------*/
BOOL TPlayer::IsOnLeftSideOfField()
{
	return ((m_iTeam == LEFT_TO_RIGHT) && (GetAbsY() < 0.0))
				|| ((m_iTeam == RIGHT_TO_LEFT) && (GetAbsY() > 0.0));
}

BOOL TPlayer::IsInOwnHalfOfField()
{
	return ((m_iTeam == LEFT_TO_RIGHT) && (GetAbsX() < 0.0))
				|| ((m_iTeam == RIGHT_TO_LEFT) && (GetAbsX() > 0.0));
}

/*-----------------------------------------------------------------------------
Function: BOOL TPlayer::IsAlife(int time)
Date    : 12.07.99
-------------------------------------------------------------------------------
Parameter: current (sensebody) time
Returnval: true, if we heard something from the player within few time
Description: checks if the specified player did say something recently
				 This only works for own players
				 Players may be too far to hear and nevertheless alife, so only
				 use for players close to this agent
-----------------------------------------------------------------------------*/
BOOL TPlayer::IsAlife(int time)
{
	// true if the player did not say something 10 times
	if (IsOwnTeam() && (time - m_iHearTimePlayer < 220))
		return TRUE;
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void	InitFreePassArea(TDirection iSecureBoundary)
Date    : 24.04.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: secure boundary in which the ball flies safe
Returnval: (void)
Description: initialize the two bordes at max values
-----------------------------------------------------------------------------*/
void TPlayer::InitFreePassArea(int iSecureBoundary)
{
	m_iFreeAreaBordersLeft=m_iDirection-iSecureBoundary;
	m_iFreeAreaBordersRight=m_iDirection+iSecureBoundary;
}

/*-----------------------------------------------------------------------------
Function: void	UpdateFreePassArea(TPlayer *pOpponent)
Date    : 24.04.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: Pointer TPlayer who covers the player
			secure boundary angle by which it is save
Returnval: (void)
Description: calculates the free pass area by the opponent who blocks
				its ensured that the player is blocked by that opponent
-----------------------------------------------------------------------------*/
//void TPlayer::UpdateFreePassArea(TPlayer* pOpponent, int iSecureBoundary)
void TPlayer::UpdateFreePassArea(TDirection pDir, int iSecureBoundary)
{
	if(m_iFreeAreaBordersLeft < m_iFreeAreaBordersRight)
	{
		if (pDir < this->GetDirection())
			m_iFreeAreaBordersLeft = pDir + iSecureBoundary;
		else
			m_iFreeAreaBordersRight = pDir - iSecureBoundary;

		if (m_iFreeAreaBordersLeft > m_iFreeAreaBordersRight)
			m_iFreeAreaBordersLeft = m_iFreeAreaBordersRight;
	}
}

/*-----------------------------------------------------------------------------
Function: int	GetFreePassArea()
Date    : 24.04.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: none
Returnval: int
Description: gives the range of the free pass area
-----------------------------------------------------------------------------*/
int	TPlayer::GetFreePassArea()
{
	return  m_iFreeAreaBordersRight - m_iFreeAreaBordersLeft;
}

/*-----------------------------------------------------------------------------
Function: int	GetFreePassDirection()
Date    : 24.04.99
From    : Markus Plewinski
-------------------------------------------------------------------------------
Parameter: none
Returnval: int
Description: gives the angle of the free pass direction
-----------------------------------------------------------------------------*/
int	TPlayer::GetFreePassDirection()
{
	return m_iFreeAreaBordersLeft + (GetFreePassArea() / 2);
}


/*-----------------------------------------------------------------------------
Function: void TPlayer::GetFuturePositionActive(double timeAhead, int dashpower, int direction
												  , double effort, TVector& pos)
Date    : 24.06.99
-------------------------------------------------------------------------------
Parameter: timesteps in future to calculate the positions of this object
			  power the player is expected to dash
			  absolute direction the player dashes
			  effort of the player from sense_body
			  vector to store the result into
Returnval: (void)
Description: calculates the expected position of a player that is expected to
				 constantly dash with dashpower into current direction
-----------------------------------------------------------------------------*/
void TPlayer::GetFuturePositionActive(double timeAhead, int dashpower, int direction
												  , double effort, TVector& pos)
{
	ASSERT(timeAhead >= 0);
	
	int		i;
	double	dAcc = DASH_POWER_RATE * effort * dashpower;

	m_tFutureSpeed = m_tAbsSpeed;
	m_tFutureAcc.SetVectorByDirection(dAcc, direction);
	m_tFuturePos = m_tPosition;
	
	for (i = 0; i < timeAhead; i++)
		GetNextFuturePosition();

	pos = m_tFuturePos;
}

void TPlayer::GetNextFuturePosition()
{
	m_tFutureSpeed += m_tFutureAcc;
	m_tFuturePos += m_tFutureSpeed;
	m_tFutureSpeed *= GetDecay();
}

/*-----------------------------------------------------------------------------
Function: double TPlayer::GetFutureDistanceActive(double timeAhead, int dashpower, int dir, double effort)
Date    : 24.06.99
-------------------------------------------------------------------------------
Parameter: timesteps in future to calculate the positions of this object
			  power the player is expected to dash
			  direction the player is expected to dash
			  effort of the player from sense_body
Returnval: distance the player can go within the specified time
Description: calculates the expected distance a player can go that is expected to
				 constantly dash with dashpower
-----------------------------------------------------------------------------*/
double TPlayer::GetFutureDistanceActive(double timeAhead, int dashpower, int dir, double effort)
{
	ASSERT(timeAhead >= 0);
	
	int		i;
	double	dAcc = DASH_POWER_RATE * effort * dashpower;
	double	ret = 0.0;

	m_tFutureSpeed = m_tAbsSpeed;
	m_tFutureAcc.SetVectorByDirection(dAcc, dir);
	m_tFuturePos = TVector(0.0, 0.0);
	
	for (i = 0; i < timeAhead; i++)
		ret = GetNextFutureDistance();
	
	return ret;
}

double TPlayer::GetNextFutureDistance()
{
	m_tFutureSpeed += m_tFutureAcc;
	m_tFuturePos += m_tFutureSpeed;
	m_tFutureSpeed *= GetDecay();
	return m_tFuturePos.GetLength();
}

/*-----------------------------------------------------------------------------
Function: void TBall::SetBall(int time, double dist, int dir, double distchng, int dirchng)
Date    : 01.04.98
-------------------------------------------------------------------------------
Parameter: parameters of the ball
Returnval: (void)
Description: sets the balls parameters
-----------------------------------------------------------------------------*/
void TBall::SetBall(int time, double dist, int dir, double distchng, double dirchng)
{
	SetMoveableObject(time, dist, dir, distchng, dirchng);
}

/*-----------------------------------------------------------------------------
Function: void TBall::DrawObject(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
			  reference to the sreen paramters (mid, scale)
Returnval: (void)
Description: draws the ball as a circle 
-----------------------------------------------------------------------------*/
void TBall::DrawObject(CDC* pDC, TScreenParams *tScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	int		size;
	int		age;

	// draw only players having been percepted this cycle
	if (!tScreen->IsAbsoluteView() && !IsValid())
		return;
	if (tScreen->IsAbsoluteView() && !GetOnceSeen())
		return;

	// draw a Ellipse at the players position
	if (tScreen->IsTopView())
	{
		size = BALLSIZE;
		cCol = COLOR_BALL;
	}
	else if (tScreen->IsRelativeView())
	{
		if (m_dDistance > EYE_HEIGHT)
			size = (int) (BALLSIZE  * SCALE_DIST / m_dDistance);
		else
			size = (int) (BALLSIZE  * SCALE_DIST / EYE_HEIGHT);
		cCol = COLOR_BALL;
	}
	else
	{
		// absolute view
		if (!IsInMind())
			return;

		// color of the ball is fading when it has been seen long time ago
		age = tScreen->GetTime() - m_iTime;
		if (age < -7)
			age = -7;
		cCol = RGB(20, 20, 200 - age * 200 /MAX_IMAGINE_TIME);
		size = (int) (BALL_SIZE_DRAW * tScreen->GetAbsScale());
	}

	// draw the circle
	brush.CreateSolidBrush(cCol);
	oldBrush = pDC->SelectObject(&brush);
	SetCoordinates(tScreen);

	pDC->Ellipse(m_iX - size, m_iY - size, m_iX + size, m_iY + size);
	
	pDC->SelectObject(oldBrush);
	brush.DeleteObject();
}


TSoccerGoal::TSoccerGoal(double x, double y)
{
	SetPosition(TVector(x, y));
}

/*-----------------------------------------------------------------------------
Function: void TSoccerGoal::DrawObject(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
			  reference to the sreen paramters (mid, scale)
Returnval: (void)
Description: draws the goal as a black box
-----------------------------------------------------------------------------*/
void TSoccerGoal::DrawObject(CDC* pDC, TScreenParams *tScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	int		size;

	// draw only players having been percepted this cycle
	if (!IsValid())
		return;

	if (tScreen->IsAbsoluteView())
		// the goal in abs view is drawn by DrawVisibleArea()
		return;

	// color of the goal
	cCol = COLOR_GOAL;

	// draw the polygone
	brush.CreateSolidBrush(cCol);
	oldBrush = pDC->SelectObject(&brush);
	SetCoordinates(tScreen);

	// draw a rectangle
	if (tScreen->IsTopView())
		size = GOALSIZE;
	else if (tScreen->IsRelativeView())
	{
		if (m_dDistance > EYE_HEIGHT)
			size = (int) (GOALSIZE  * SCALE_DIST / m_dDistance);
		else
			size = (int) (GOALSIZE  * SCALE_DIST / EYE_HEIGHT);
	}

	pDC->Rectangle(m_iX - size - size, m_iY - size, m_iX + size + size, m_iY + size);
	
	pDC->SelectObject(oldBrush);
	brush.DeleteObject();
}

// additionally draws the distance of the goal to the player in top view
void TSoccerGoal::DrawGoal(CDC* pDC, TScreenParams *tScreen, double dist)
{
	DrawObject(pDC, tScreen);

	// to test the distance to the goals
	if (tScreen->GetDrawDegreeOfFreeness())
	{
		int		nPointSize;
		CFont		font;
		CFont*	oldFont;
		CRect		field;
		COLORREF	oldTextColor;
		COLORREF	oldBkColor;
		int		sizeX;
		CString	text;

		nPointSize = (int) (20.0 * tScreen->GetAbsScale());
		if (!font.CreatePointFont(nPointSize, _T("Times New Roman"), pDC))
			font.CreateStockObject(SYSTEM_FONT);
		oldFont = pDC->SelectObject(&font);

		// calculate the rectangle of the text
		sizeX = (int) tScreen->GetScale() * nPointSize / 5;
		oldTextColor = pDC->SetTextColor(0x00000000);
		oldBkColor = pDC->SetBkColor(0xffffffff);

		// draw distance to this agent
		field.SetRect(m_iX, m_iY, m_iX + sizeX, m_iY + nPointSize / 5);
		text.Format("%5.2f", dist);
		pDC->DrawText(text, field, DT_SINGLELINE);

		pDC->SetTextColor(oldTextColor);
		pDC->SetBkColor(oldBkColor);
		pDC->SelectObject(oldFont);
		font.DeleteObject();
	}
}

/*-----------------------------------------------------------------------------
Function: TFlag::TFlag(double x, double y)
Date    : 18.12.97
-------------------------------------------------------------------------------
Parameter: none
Description: a flag is by default a field flag
-----------------------------------------------------------------------------*/
TFlag::TFlag(double x, double y)
{
	SetPosition(TVector(x, y));
	m_iType = FIELD_FLAG;
}

/*-----------------------------------------------------------------------------
Function: _
Date    : 14.11.97
-------------------------------------------------------------------------------
Parameter: _
Returnval: (void)
Description: _
-----------------------------------------------------------------------------*/
void TFlag::SetFlag(int type, char tag1, char tag2)
{
	m_iType = type;
	m_cTag1 = tag1;
	m_cTag2 = tag2;
}

/*-----------------------------------------------------------------------------
Function: void TFlag::DrawObject(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
			  reference to the sreen paramters (mid, scale)
Returnval: (void)
Description: draws the flag as a triangle
-----------------------------------------------------------------------------*/
void TFlag::DrawObject(CDC* pDC, TScreenParams *tScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	CPoint	point[3];
	int		size;

	// draw only players having been percepted this cycle
	if (!IsValid())
		return;

	// color of the ball
	cCol = COLOR_FLAG;

	// draw the polygone
	brush.CreateSolidBrush(cCol);
	oldBrush = pDC->SelectObject(&brush);
	SetCoordinates(tScreen);

	// draw a Ellipse at the players position
	if (tScreen->IsTopView())
		size = FLAG_SIZE;
	else if (tScreen->IsRelativeView())
	{
		if (m_dDistance > EYE_HEIGHT)
			size = (int) (FLAG_SIZE  * SCALE_DIST / m_dDistance);
		else
			size = (int) (FLAG_SIZE  * SCALE_DIST / EYE_HEIGHT);
	}
	else
		// absolute view
		size = FLAG_SIZE;

	if (m_iType == FIELD_FLAG)
	{
		point[0].x = m_iX - size;
		point[0].y = m_iY - size;
		point[1].x = m_iX + size;
		point[1].y = m_iY;
		point[2].x = m_iX - size;
		point[2].y = m_iY + size;
	}
	else
	{
		point[0].x = m_iX + size;
		point[0].y = m_iY - size;
		point[1].x = m_iX - size;
		point[1].y = m_iY;
		point[2].x = m_iX + size;
		point[2].y = m_iY + size;
	}
	pDC->Polygon(point, 3);

	pDC->SelectObject(oldBrush);
	brush.DeleteObject();
}

/*-----------------------------------------------------------------------------
Function: void	TFlag::DrawLine(TFlag* pFlag, CDC *pDC, TScreenParams *tScreen)
Date    : 14.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the flag to which the line is drawn
			  pointer to the device context on which should be drawn
			  pointer to the screen parameters for scaling and translation
Returnval: (void)
Description: draws a line from this flag to the passed flag, if both flags are
				 seen from this point of view
-----------------------------------------------------------------------------*/
void	TFlag::DrawLine(TFlag* pFlag, CDC *pDC, TScreenParams *tScreen)
{
	CPen	pen;
	CPen* oldPen;
	COLORREF cCol = COLOR_LINE;

	// test if the two flags are valid
	if (!IsValid() || !pFlag->IsValid())
		return;

	// prepare the pen
	pen.CreatePen(PS_SOLID, 1, cCol);
	oldPen = pDC->SelectObject(&pen);

	// get the proper coordinates
	pDC->MoveTo(m_iX, m_iY);
	pDC->LineTo(pFlag->GetX(), pFlag->GetY());

	pDC->SelectObject(oldPen);
	pen.DeleteObject();
}

/*-----------------------------------------------------------------------------
Function: void TLine::DrawObject(CDC* pDC, TScreenParams *tScreen)
Date    : 18.12.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
			  reference to the sreen paramters (mid, scale)
Returnval: (void)
Description: draws the flag as a triangle
-----------------------------------------------------------------------------*/
void TLine::DrawObject(CDC* pDC, TScreenParams *tScreen)
{
	COLORREF cCol;
	CBrush	brush;
	CBrush	*oldBrush;
	CPoint	point[4];

	// draw only players having been percepted this cycle
	if (!IsValid())
		return;

	// color of the ball
	cCol = COLOR_LINE;

	// draw the polygone
	brush.CreateSolidBrush(cCol);
	oldBrush = pDC->SelectObject(&brush);
	SetCoordinates(tScreen);

	// draw a raute at the lines position
	point[0].x = m_iX;
	point[0].y = m_iY - FLAG_SIZE;
	point[1].x = m_iX + FLAG_SIZE;
	point[1].y = m_iY;
	point[2].x = m_iX;
	point[2].y = m_iY + FLAG_SIZE;
	point[3].x = m_iX - FLAG_SIZE;
	point[3].y = m_iY;
	pDC->Polygon(point, 4);

	pDC->SelectObject(oldBrush);
	brush.DeleteObject();
}
